from __future__ import annotations
from abc import abstractmethod
import jax.numpy as np
from jax import Array
from jax.scipy.stats import norm


from ..psfs import PSF
from .optical_layers import BaseLayer

__all__ = [
    "ApplyPixelResponse",
    "ApplyJitter",
    "ApplySaturation",
    "AddConstant",
    "Downsample",
]


class DetectorLayer(BaseLayer):
    """
    A base Detector layer class to help with type checking throughout the rest of the
    software.
    """

    def __init__(self: DetectorLayer):
        super().__init__()

    @abstractmethod
    def apply(self: DetectorLayer, psf: PSF) -> PSF:  # pragma: no cover
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """


class ApplyPixelResponse(DetectorLayer):
    """
    Applies a pixel response array to the input psf, via a multiplication. This can be
    used to model variations in the inter and intra-pixel sensitivity variations common
    to most detectors.

    ??? abstract "UML"
        ![UML](../../assets/uml/ApplyPixelResponse.png)

    Attributes
    ----------
    pixel_response : Array
        The pixel_response to apply to the input psf.
    """

    pixel_response: Array

    def __init__(self: DetectorLayer, pixel_response: Array):
        """
        Parameters
        ----------
        pixel_response : Array
            The pixel_response to apply to the input psf. Must be a 2-dimensional array
            equal to size of the psf at time of application.
        """
        super().__init__()
        self.pixel_response = np.asarray(pixel_response, dtype=float)
        if self.pixel_response.ndim != 2:
            raise ValueError("pixel_response must be a 2 dimensional array.")

    def apply(self: DetectorLayer, psf: PSF) -> PSF:
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """
        return psf * self.pixel_response


class ApplyJitter(DetectorLayer):
    """
    Convolves the psf with a radially symmetric Gaussian kernel parameterised by its
    standard deviation (sigma).

    ??? abstract "UML"
        ![UML](../../assets/uml/ApplyJitter.png)

    Attributes
    ----------
    sigma : float, pixels
        The standard deviation of the Gaussian kernel, in units of pixels.
    kernel_size : int
        The size of the convolution kernel to use.
    """

    kernel_size: int
    sigma: float

    def __init__(self: DetectorLayer, sigma: float, kernel_size: int = 10):
        """
        Parameters
        ----------
        sigma : float, pixels
            The standard deviation of the Gaussian kernel, in units of pixels.
        kernel_size : int = 10
            The size of the convolution kernel to use.
        """
        super().__init__()
        self.kernel_size = int(kernel_size)
        self.sigma = float(sigma)

    def generate_kernel(self: DetectorLayer, pixel_scale: float) -> Array:
        """
        Generates the normalised Gaussian kernel.

        Returns
        -------
        kernel : Array
            The Gaussian kernel.
        """
        # TODO: Move to utils?
        # Generate distribution
        sigma = self.sigma * pixel_scale
        x = np.linspace(-10, 10, self.kernel_size) * pixel_scale
        kernel = norm.pdf(x, scale=sigma) * norm.pdf(x[:, None], scale=sigma)
        return kernel / np.sum(kernel)

    def apply(self: DetectorLayer, psf: PSF) -> PSF:
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """
        kernel = self.generate_kernel(psf.pixel_scale)
        return psf.convolve(kernel)


class ApplySaturation(DetectorLayer):
    """
    Applies a simple saturation model to the input psf, by clipping any values above
    the threshold value.

    ??? abstract "UML"
        ![UML](../../assets/uml/ApplySaturation.png)

    Attributes
    ----------
    threshold : float
        The threshold at which the saturation is applied.
    """

    threshold: float

    def __init__(self: DetectorLayer, threshold: float):
        """
        Parameters
        ----------
        threshold : float
            The threshold at which the saturation is applied.
        """
        super().__init__()
        self.threshold = float(threshold)

    def apply(self: DetectorLayer, psf: PSF) -> PSF:
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """
        return psf.min("data", self.threshold)


class AddConstant(DetectorLayer):
    """
    Adds a constant to the output psf. This is typically used to model the mean value of
    the detector noise.

    ??? abstract "UML"
        ![UML](../../assets/uml/AddConstant.png)

    Attributes
    ----------
    value : float
        The value to add to the psf.
    """

    value: float

    def __init__(self: DetectorLayer, value: float):
        """
        Parameters
        ----------
        value : float
            The value to add to the psf.
        """
        super().__init__()
        self.value = float(value)

    def apply(self: DetectorLayer, psf: PSF) -> PSF:
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """
        return psf + self.value


class Downsample(DetectorLayer):
    """
    Downsamples an input psf by an integer number of pixels via a sum. Typically used
    to downsample an oversampled psf to the true pixel size. Note kernel_size must be
    an integer multiple of the input psf size.

    ??? abstract "UML"
        ![UML](../../assets/uml/Downsample.png)

    Attributes
    ----------
    kernel_size : int
        The size of the downsampling kernel.
    """

    kernel_size: int

    def __init__(self: DetectorLayer, kernel_size: int):
        """
        Parameters
        ----------
        kernel_size : int
            The size of the downsampling kernel.
        """
        super().__init__()
        self.kernel_size = int(kernel_size)

    def apply(self: DetectorLayer, psf: PSF) -> PSF:
        """
        Applies the layer to the PSF.

        Parameters
        ----------
        psf : PSF
            The psf to operate on.

        Returns
        -------
        psf : PSF
            The transformed psf.
        """
        return psf.downsample(self.kernel_size)



from __future__ import annotations
from typing import Any
from abc import abstractmethod
import jax.numpy as np
import jax.tree as jtu
from jax.scipy.signal import convolve
from jax import vmap, Array, tree
from zodiax import filter_vmap, Base
import dLux.utils as dlu
from dLux import spectra
import dLux


from .psfs import PSF

Optics = lambda: dLux.optical_systems.BaseOpticalSystem
Spectrum = lambda: dLux.spectra.BaseSpectrum


__all__ = [
    "BaseSource",
    "PointSource",
    "PointSources",
    "BinarySource",
    "ResolvedSource",
    "PointResolvedSource",
    "Scene",
]


class BaseSource(Base):
    @abstractmethod
    def normalise(self):  # pragma: no cover
        pass

    @abstractmethod
    def model(self, optics):  # pragma: no cover
        pass


class Source(BaseSource):
    """
    Base source class that implements the spectra attribute.

    Attributes
    ----------
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    spectrum: Spectrum()

    def __init__(
        self: Source,
        wavelengths: Array = None,
        weights: Array = None,
        spectrum: Spectrum() = None,
    ):
        """
        Parameters
        ----------
        wavelengths : Array, metres = None
            The array of wavelengths at which the spectrum is defined. This input is
            ignored if a Spectrum object is provided.
        weights : Array = None
            The spectral weights of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        # Spectrum
        if spectrum is not None:
            if not isinstance(spectrum, spectra.Spectrum):
                raise TypeError("spectrum must be a dLux Spectrum object.")
            self.spectrum = spectrum
        else:
            self.spectrum = spectra.Spectrum(wavelengths, weights)

    def __getattr__(self: Source, key: str) -> Any:
        """
        Raises the parameters of the spectrum object to this class.

        Parameters
        ----------
        key : str
            The key to get from the spectrum object.

        Returns
        -------
        value : Any
            The value of the key.
        """
        if hasattr(self.spectrum, key):
            return getattr(self.spectrum, key)
        else:
            raise AttributeError(
                f"{self.__class__.__name__} has no " f"attribute {key}."
            )

    def normalise(self: Source) -> Source:
        """
        Returns a normalised source object.

        Returns
        -------
        source : Source
            The normalised source object.
        """
        norm_spectrum = self.spectrum.normalise()
        return self.set("spectrum", norm_spectrum)


class PointSource(Source):
    """
    A simple point source with a spectrum, position and flux.

    ??? abstract "UML"
        ![UML](../../assets/uml/PointSource.png)

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : float, photons
        The flux of the object.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    position: Array
    flux: float

    def __init__(
        self: Source,
        wavelengths: Array = None,
        position: Array = np.zeros(2),
        flux: float = 1.0,
        weights: Array = None,
        spectrum: Spectrum() = None,
    ):
        """
        Parameters
        ----------
        wavelengths : Array, metres = None
            The array of wavelengths at which the spectrum is defined. This input is
            ignored if a Spectrum object is provided.
        position : Array, radians = np.zeros(2)
            The (x, y) on-sky position of this object.
        flux : float, photons = 1.
            The flux of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        # Position and Flux
        self.position = np.asarray(position, dtype=float)
        self.flux = float(flux)

        if self.position.shape != (2,):
            raise ValueError("position must be a 1d array of shape (2,).")

        super().__init__(
            wavelengths=wavelengths, weights=weights, spectrum=spectrum
        )

    def model(
        self: Source,
        optics: Optics,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        self = self.normalise()
        weights = self.weights * self.flux
        return optics.propagate(
            self.wavelengths, self.position, weights, return_wf, return_psf
        )


class PointSources(Source):
    """
    A set of point sources with the same spectrum, but different positions and fluxes.

    ??? abstract "UML"
        ![UML](../../assets/uml/PointSources.png)

    Attributes
    ----------
    position : Array, radians
        The ((x0, y0), (x1, y1), ...) on-sky positions of these sources.
    flux : Array, photons
        The fluxes of the sources.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    position: Array
    flux: Array

    def __init__(
        self: Source,
        wavelengths: Array = None,
        position: Array = np.zeros((1, 2)),
        flux: Array = None,
        weights: Array = None,
        spectrum: Spectrum() = None,
    ):
        """
        Parameters
        ----------
        wavelengths : Array, metres
            The array of wavelengths at which the spectrum is defined.
        position : Array, radians = np.zeros((1, 2))
            The (x, y) on-sky position of this object.
        flux : Array, photons = None
            The flux of the object.
        weights : Array = None
            The spectral weights of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        super().__init__(
            spectrum=spectrum, wavelengths=wavelengths, weights=weights
        )

        # More complex parameter checks here because of extra dims
        self.position = np.asarray(position, dtype=float)
        if self.position.ndim != 2:
            raise ValueError("position must be a 2d array.")

        if flux is None:
            self.flux = np.ones(len(self.position))
        else:
            self.flux = np.asarray(flux, dtype=float)

            if self.flux.ndim != 1:
                raise ValueError("flux must be a 1d array.")

            if len(self.flux) != len(self.position):
                raise ValueError(
                    "Length of flux must be equal to length of " "positions."
                )

    def model(
        self: Source,
        optics: Optics,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        if return_wf and return_psf:
            raise ValueError(
                "return_wf and return_psf cannot both be True. "
                "Please choose one."
            )
        self = self.normalise()
        weights = self.weights[None, :] * self.flux[:, None]
        prop_fn = lambda position, weight: optics.propagate(
            self.wavelengths, position, weight, return_wf=True
        )
        wfs = filter_vmap(prop_fn)(self.position, weights)

        if return_wf:
            return wfs
        if return_psf:
            return PSF(wfs.psf.sum((0, 1)), wfs.pixel_scale.mean())
        else:
            return wfs.psf.sum((0, 1))


class ResolvedSource(PointSource):
    """
    A single resolved source with a spectrum, position, flux, and distribution array
    that represents the resolved component.

    ??? abstract "UML"
        ![UML](../../assets/uml/ResolvedSource.png)

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : float, photons
        The flux of the object.
    distribution : Array
        The array of intensities representing the resolved source.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    distribution: Array

    def __init__(
        self: Source,
        wavelengths: Array = None,
        position: Array = np.zeros(2),
        flux: float = 1.0,
        distribution: Array = np.ones((3, 3)),
        weights: Array = None,
        spectrum: Spectrum() = None,
    ):
        """
        Parameters
        ----------
        wavelengths : Array, metres
            The array of wavelengths at which the spectrum is defined.
        position : Array, radians = np.zeros(2)
            The (x, y) on-sky position of this object.
        flux : float, photons = 1.
            The flux of the object.
        distribution : Array = np.ones((3, 3))
            The array of intensities representing the resolved source.
        weights : Array = None
            The spectral weights of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        distribution = np.asarray(distribution, dtype=float)
        self.distribution = distribution / distribution.sum()

        if self.distribution.ndim != 2:
            raise ValueError("distribution must be a 2d array.")

        super().__init__(
            position=position,
            flux=flux,
            spectrum=spectrum,
            wavelengths=wavelengths,
            weights=weights,
        )

    def normalise(self: Source) -> Source:
        """
        Method for returning a new source object with a normalised total
        spectrum and source distribution.

        Returns
        -------
        source : Source
            The source object with the normalised spectrum and distribution.
        """
        spectrum = self.spectrum.normalise()
        distribution_floor = np.maximum(self.distribution, 0.0)
        distribution = distribution_floor / distribution_floor.sum()
        return self.set(["spectrum", "distribution"], [spectrum, distribution])

    def model(
        self: Source,
        optics: Optics = None,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        if return_wf and return_psf:
            raise ValueError(
                "return_wf and return_psf cannot both be True. "
                "Please choose one."
            )
        # Normalise and get parameters
        self = self.normalise()
        weights = self.weights * self.flux

        # Note we always return wf here so we can convolve each wavelength
        # individually if a chromatic wavefront output is required.
        wf = optics.propagate(
            self.wavelengths, self.position, weights, return_wf=True
        )

        # Returning wf is a special case
        if return_wf:
            conv_fn = lambda psf: convolve(psf, self.distribution, mode="same")
            return wf.set("amplitude", vmap(conv_fn)(wf.psf) ** 0.5)

        # Return psf object
        conv_psf = convolve(wf.psf.sum(0), self.distribution, mode="same")
        if return_psf:
            return PSF(conv_psf, wf.pixel_scale.mean())

        # Return array psf
        return conv_psf


class BinarySource(Source):
    """
    A binary source parameterised by the position, flux, separation, position_angle,
    and contrast between the two sources.

    ??? abstract "UML"
        ![UML](../../assets/uml/BinarySource.png)

    Attributes
    ----------
    position : Array, radians
        The mean (x, y) on-sky position of this object.
    mean_flux : float, photons
        The mean flux of the sources.
    separation : float, radians
        The separation of the two sources in radians.
    position_angle : float, radians
        The position angle between the two sources measured clockwise from the
        vertical axis.
    contrast : float
        The contrast ratio between the two sources.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    position: Array
    mean_flux: float
    separation: float
    position_angle: float
    contrast: float

    def __init__(
        self: Source,
        wavelengths: Array = None,
        position: Array = np.zeros(2),
        mean_flux: float = 1.0,
        separation: float = 0.0,
        position_angle: float = np.pi / 2,
        contrast: float = 1.0,
        spectrum: Spectrum() = None,
        weights: Array = None,
    ):
        """
        Parameters
        ----------
        wavelengths : Array, metres = None
            The array of wavelengths at which the spectrum is defined.
        position : Array, radians = np.zeros(2)
            The (x, y) on-sky position of this object.
        mean_flux : float, photons = 1.
            The mean flux of the sources.
        separation : float, radians = 0.
            The separation of the two sources in radians.
        position_angle : float, radians = np.pi / 2
            The position angle between the two sources measured clockwise from the
            vertical axis.
        contrast : float = 1.
            The contrast ratio between the two sources.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        wavelengths = np.asarray(wavelengths, dtype=float)
        if weights is None:
            weights = np.ones((2, len(wavelengths)))

        # Position and Flux
        self.position = np.asarray(position, dtype=float)
        self.mean_flux = float(mean_flux)

        if self.position.shape != (2,):
            raise ValueError("position must be a 1d array of shape (2,).")

        # Binary values
        self.separation = float(separation)
        self.position_angle = float(position_angle)
        self.contrast = float(contrast)

        super().__init__(
            wavelengths=wavelengths,
            spectrum=spectrum,
            weights=weights,
        )

    def model(
        self: Source,
        optics: Optics,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        # Normalise and get input values
        self = self.normalise()
        positions = dlu.positions_from_sep(
            self.position, self.separation, self.position_angle
        )
        flux = dlu.fluxes_from_contrast(self.mean_flux, self.contrast)
        weights = self.weights * flux[:, None]

        # Return wf case is simple
        prop_fn = lambda position, weight: optics.propagate(
            self.wavelengths, position, weight, return_wf, return_psf
        )
        output = filter_vmap(prop_fn)(positions, weights)

        # Return wf is simple case
        if return_wf:
            return output

        # Return psf just requires constructing object
        if return_psf:
            return PSF(output.data.sum(0), output.pixel_scale.mean())

        # Return array is simple
        return output.sum(0)


class PointResolvedSource(ResolvedSource):
    """
    A class for modelling a point source and a resolved source that is defined
    relative to the point source. An example would be an unresolved star with
    a resolved dust shell or debris disk. These two objects share the same
    spectra but have their flux defined by flux (the mean flux) and the flux
    ratio (contrast) between the point source and resolved distribution. The
    resolved component is defined by an array of intensities that represent
    the resolved distribution.

    ??? abstract "UML"
        ![UML](../../assets/uml/PointResolvedSource.png)

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : float, photons
        The mean flux of the point and resolved source.
    distribution : Array
        The array of intensities representing the resolved source.
    contrast : float
        The contrast ratio between the point source and the resolved source.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    """

    contrast: float

    def __init__(
        self: Source,
        wavelengths: Array = None,
        position: Array = np.zeros(2),
        flux: float = 1.0,
        distribution: Array = np.ones((3, 3)),
        contrast: float = 1.0,
        weights: Array = None,
        spectrum: Spectrum() = None,
    ) -> Source:
        """
        Parameters
        ----------
        wavelengths : Array, metres = None
            The array of wavelengths at which the spectrum is defined.
        position : Array, radians = np.zeros(2)
            The (x, y) on-sky position of this object.
        flux : float, photons = 1.
            The mean flux of the point and resolved source.
        distribution : Array = np.ones((3, 3))
            The array of intensities representing the resolved source.
        contrast : float = 1.
            The contrast ratio between the point source and the resolved source.
        weights : Array = None
            The spectral weights of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        """
        wavelengths = np.asarray(wavelengths, dtype=float)
        if weights is None:
            weights = np.ones((2, len(wavelengths)))

        self.contrast = float(contrast)

        super().__init__(
            wavelengths=wavelengths,
            position=position,
            flux=flux,
            distribution=distribution,
            spectrum=spectrum,
            weights=weights,
            # contrast=contrast,
        )

    def model(
        self: Source,
        optics: Optics,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        if return_wf and return_psf:
            raise ValueError(
                "return_wf and return_psf cannot both be True. "
                "Please choose one."
            )
        # Normalise and get parameters
        self = self.normalise()
        flux = dlu.fluxes_from_contrast(self.flux, self.contrast)
        weights = self.weights * flux[:, None]

        # Note we always return wf here so we can convolve each wavelength
        # individually if a chromatic wavefront output is required. We also
        # Can not propagate the weights since they have different values
        # for the point and resolved source.
        wf = optics.propagate(self.wavelengths, self.position, return_wf=True)

        # Returning wf is a special case, we need to convolve each psf with
        # the distribution, and them re-combine them into a vectorised wf
        if return_wf:
            # Perform convolution
            conv_fn = lambda psf: convolve(psf, self.distribution, mode="same")
            conv_wf = wf.set("amplitude", vmap(conv_fn)(wf.psf) ** 0.5)

            # Stack leaves manually, this is a bit of a hack to get around
            # string leaf errors from jtu.map, and to avoid
            # flattening/unflattening with partition and combine
            stack_leaves = lambda x, y: np.stack([x, y], axis=0)
            amplitudes = stack_leaves(wf.amplitude, conv_wf.amplitude)
            phases = stack_leaves(wf.phase, conv_wf.phase)
            pixel_scales = stack_leaves(wf.pixel_scale, conv_wf.pixel_scale)
            wavelengths = stack_leaves(wf.wavelength, conv_wf.wavelength)

            # Combine into single wf and finally apply weights
            combined_wf = wf.set(
                ["wavelength", "amplitude", "phase", "pixel_scale"],
                [wavelengths, amplitudes, phases, pixel_scales],
            )
            return combined_wf.multiply("amplitude", weights[:, :, None, None])

        # Create singe array psf object
        point_psf = (np.expand_dims(weights[0], (1, 2)) * wf.psf).sum(0)
        resolved_psf = (np.expand_dims(weights[1], (1, 2)) * wf.psf).sum(0)
        conv_psf = convolve(resolved_psf, self.distribution, mode="same")
        psf = point_psf + conv_psf
        if return_psf:
            return PSF(psf, wf.pixel_scale.mean())

        # Return array psf
        return psf


class Scene(BaseSource):
    """
    A source object that holds a set of sources that are model simultaneously.

    ??? abstract "UML"
        ![UML](../../assets/uml/Scene.png)

    Attributes
    ----------
    sources : dict
        A dictionary of source objects to model simultaneously.
    """

    sources: dict

    def __init__(self: Scene, sources: list[Source]):
        """
        Parameters
        ----------
        sources : list[Source]
            A list of source objects to model simultaneously.
        """
        super().__init__()
        if isinstance(sources, (BaseSource, tuple)):
            sources = [sources]
        self.sources = dlu.list2dictionary(sources, False, BaseSource)

    def normalise(self: Scene) -> Scene:
        """
        Method for returning a new scene with normalised source objects.

        Returns
        -------
        scene : Scene
            The normalised scene object.
        """
        is_source = lambda leaf: isinstance(leaf, BaseSource)
        norm_fn = lambda source: source.normalise()
        sources = jtu.map(norm_fn, self.sources, is_leaf=is_source)
        return self.set("sources", sources)

    def __getattr__(self: Source, key: str) -> Any:
        """
        Raises the individual sources via their keys.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the sub-dictionaries.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the sub-dictionaries.
        """
        if key in self.sources.keys():
            return self.sources[key]
        raise AttributeError(
            f"{self.__class__.__name__} has no attribute " f"{key}."
        )

    def model(
        self: Scene,
        optics: Optics(),
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the source object through the provided optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source object.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        self = self.normalise()

        # Define leaf_fn and map across sources
        leaf_fn = lambda leaf: isinstance(leaf, BaseSource)
        output = jtu.map(
            lambda s: s.model(optics, return_wf, return_psf),
            self.sources,
            is_leaf=leaf_fn,
        )

        # Return wf case is simple
        if return_wf:
            return output

        # Return psf case requires mapping across the psf outputs
        if return_psf:
            # Define mapping function
            leaf_fn = lambda leaf: isinstance(leaf, PSF)
            get_psfs = lambda psf: psf.data.sum(tuple(range(psf.ndim)))
            get_pscales = lambda psf: psf.pixel_scale.mean()

            # Get values and return PSF
            psf = dlu.map2array(get_psfs, output, leaf_fn).sum(0)
            pixel_scale = dlu.map2array(get_pscales, output, leaf_fn).mean()
            return PSF(psf, pixel_scale)

        # Return array is simple
        return dlu.map2array(lambda x: x, output).sum(0)



from jax import lax, Array
import jax.numpy as np
from typing import Any

__all__ = ["factorial", "triangular_number", "eval_basis", "nandiv"]


# TODO: Use lax.cond to make n == 0 a safe case
def factorial(n: float) -> float:
    """
    Calculate n! in a jax friendly way. Note that n == 0 is not a
    safe case.

    Parameters
    ----------
    n : float
        The value to calculate the factorial of.

    Returns
    -------
    n! : float
        The factorial of the value.
    """
    return lax.exp(lax.lgamma(n + 1.0))


def triangular_number(n: int) -> int:
    """
    Calculate the nth triangular number.

    Parameters
    ----------
    n : int
        The nth triangular number to calculate.

    Returns
    -------
    n : int
        The nth triangular number.
    """
    return n * (n + 1) / 2


def eval_basis(basis: Array, coefficients: Array) -> Array:
    """
    Performs an n-dimensional dot-product between the basis and coefficients arrays.

    Parameters
    ----------
    basis: Array
        The basis to use.
    coefficients: Array
        The Array of coefficients to be applied to each basis vector.
    """
    ndim = coefficients.ndim
    return np.tensordot(basis, coefficients, axes=2 * (tuple(range(ndim)),))


def nandiv(a: Array, b: Array, fill: Any = np.inf) -> Array:
    """
    Divides two arrays, replacing any NaNs with a fill value.

    Parameters
    ----------
    a : Array
        The numerator.
    b : Array
        The denominator.
    fill : Any = np.inf
        The value to replace NaNs with.

    Returns
    -------
    a / b : Array
        The result of the division.
    """
    return np.where(b == 0, fill, a / b)



# Import as modules
from . import (
    propagation,
    coordinates,
    helpers,
    interpolation,
    math,
    optics,
    units,
    array_ops,
    zernikes,
    source,
    geometry,
)

# Add to __all__
modules = [
    propagation,
    coordinates,
    helpers,
    interpolation,
    units,
    optics,
    math,
    array_ops,
    zernikes,
    source,
    geometry,
]

__all__ = [module.__all__ for module in modules]

# Dont import all functions from modules
from .propagation import FFT as FFT, MFT as MFT, fresnel_MFT as fresnel_MFT
from .coordinates import (
    cart2polar as cart2polar,
    polar2cart as polar2cart,
    pixel_coords as pixel_coords,
    nd_coords as nd_coords,
    translate_coords as translate_coords,
    compress_coords as compress_coords,
    shear_coords as shear_coords,
    rotate_coords as rotate_coords,
)
from .helpers import (
    map2array as map2array,
    list2dictionary as list2dictionary,
    insert_layer as insert_layer,
    remove_layer as remove_layer,
)
from .interpolation import (
    generate_coordinates as generate_coordinates,
    scale as scale,
    rotate as rotate,
)
from .math import (
    factorial as factorial,
    triangular_number as triangular_number,
    eval_basis as eval_basis,
    nandiv as nandiv,
)
from .optics import (
    opd2phase as opd2phase,
    phase2opd as phase2opd,
    fringe_size as fringe_size,
)
from .units import (
    rad2arcsec as rad2arcsec,
    rad2deg as rad2deg,
    rad2arcmin as rad2arcmin,
    deg2rad as deg2rad,
    deg2arcmin as deg2arcmin,
    deg2arcsec as deg2arcsec,
    arcmin2rad as arcmin2rad,
    arcmin2deg as arcmin2deg,
    arcmin2arcsec as arcmin2arcsec,
    arcsec2rad as arcsec2rad,
    arcsec2deg as arcsec2deg,
    arcsec2arcmin as arcsec2arcmin,
)
from .source import (
    fluxes_from_contrast as fluxes_from_contrast,
    positions_from_sep as positions_from_sep,
)
from .array_ops import (
    pad_to as pad_to,
    crop_to as crop_to,
    resize as resize,
    downsample as downsample,
)
from .zernikes import (
    zernike_name as zernike_name,
    noll_indices as noll_indices,
    zernike_factors as zernike_factors,
    eval_radial as eval_radial,
    eval_azimuthal as eval_azimuthal,
    zernike as zernike,
    zernike_fast as zernike_fast,
    zernike_basis as zernike_basis,
    polike as polike,
    polike_fast as polike_fast,
    polike_basis as polike_basis,
)

from .geometry import (
    combine as combine,
    circle as circle,
    square as square,
    rectangle as rectangle,
    reg_polygon as reg_polygon,
    spider as spider,
    soft_circle as soft_circle,
    soft_square as soft_square,
    soft_rectangle as soft_rectangle,
    soft_reg_polygon as soft_reg_polygon,
    soft_spider as soft_spider,
)



from __future__ import annotations
from abc import abstractmethod
from typing import Union
import jax.numpy as np
from zodiax import Base
from jax import Array
import dLux.utils as dlu


from ..wavefronts import Wavefront


__all__ = [
    "BaseLayer",
    "TransmissiveLayer",
    "AberratedLayer",
    "BasisLayer",
    "Tilt",
    "Normalise",
]


class BaseLayer(Base):
    @abstractmethod
    def apply(self, wavefront):  # pragma: no cover
        pass


class OpticalLayer(BaseLayer):
    """
    The base optical layer class. Optical layer classes operate on `Wavefront` objects
    though their `apply` method, and are stored by the `OpticalSystem` classes.
    """

    @abstractmethod
    def apply(self, wavefront) -> Wavefront:  # pragma: no cover
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """


class TransmissiveLayer(OpticalLayer):
    """
    Base class to hold transmissive layers imbuing them with a transmission and
    normalise parameter.

    ??? abstract "UML"
        ![UML](../../assets/uml/TransmissiveLayer.png)

    Attributes
    ----------
    transmission: Array
        The Array of transmission values to be applied to the input wavefront.
    normalise: bool
        Whether to normalise the wavefront after passing through the optic.
    """

    transmission: Array
    normalise: bool

    def __init__(
        self: OpticalLayer,
        transmission: Array = None,
        normalise: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        transmission: Array = None
            The array of transmission values to be applied to the input wavefront.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the optic.
        """
        if transmission is not None:
            transmission = np.asarray(transmission, dtype=float)
        self.transmission = transmission
        self.normalise = bool(normalise)
        super().__init__(**kwargs)

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        wavefront *= self.transmission
        if self.normalise:
            wavefront = wavefront.normalise()
        return wavefront


class AberratedLayer(OpticalLayer):
    """
    Optical layer for holding static aberrations. Aberrations can be applied as either
    a phase or OPD, or both.

    ??? abstract "UML"
        ![UML](../../assets/uml/AberratedLayer.png)

    Attributes
    ----------
    opd : Array, metres
        The Array of OPD values to be applied to the input wavefront.
    phase : Array, radians
        The Array of phase values to be applied to the input wavefront.
    """

    opd: Array
    phase: Array

    def __init__(
        self: OpticalLayer,
        opd: Array = None,
        phase: Array = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        opd : Array, metres = None
            The Array of OPD values to be applied to the input wavefront.
        phase : Array, radians = None
            The Array of phase values to be applied to the input wavefront.
        """
        if opd is not None:
            opd = np.asarray(opd, dtype=float)
        self.opd = opd

        if phase is not None:
            phase = np.asarray(phase, dtype=float)
        self.phase = phase

        if self.opd is not None and self.phase is not None:
            if self.opd.shape != self.phase.shape:
                raise ValueError(
                    "opd and phase must have the same shape. Got "
                    f"shapes {self.opd.shape} and {self.phase.shape}."
                )
        super().__init__(**kwargs)

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        wavefront += self.opd
        wavefront = wavefront.add_phase(self.phase)
        return wavefront


class BasisLayer(OpticalLayer):
    """
    An OpticalLayer class that holds a set of basis vectors and coefficients, which are
    dot-producted at run time to produce the output. The basis can be applied as either
    a phase or OPD, by setting the `as_phase` attribute.

    ??? abstract "UML"
        ![UML](../../assets/uml/BasisLayer.png)

    Attributes
    ----------
    basis: Union[Array, list]
        The set of basis vectors. Should in generate be a 3 dimensional array.
    coefficients: Array
        The array of coefficients to be applied to each basis vector.
    as_phase: bool = False
        Whether to apply the basis as a phase or OPD. If True the output is applied as
        a phase, else it is applied as an OPD.
    """

    basis: Union[Array, list]
    coefficients: Array
    as_phase: bool

    # NOTE: We need the None basis input for aberrated apertures
    def __init__(
        self: OpticalLayer,
        basis: Array = None,
        coefficients: Array = None,
        as_phase: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        basis: Union[Array, list]
            The set of basis vectors. Should in generate be a 3 dimensional array.
        coefficients: Array
            The Array of coefficients to be applied to each basis vector.
        as_phase: bool = False
            Whether to apply the basis as a phase or OPD. If True the output is applied
            as a phase, else it is applied as an OPD.
        """
        super().__init__(**kwargs)

        if basis is not None:
            basis = np.asarray(basis, dtype=float)
            if coefficients is None:
                coefficients = np.zeros(basis.shape[:-2])
            else:
                coefficients = np.asarray(coefficients, dtype=float)
                if basis.shape[:-2] != coefficients.shape:
                    raise ValueError(
                        "The number of basis vectors must be equal to "
                        "the number of coefficients."
                    )

        self.basis = basis
        self.coefficients = coefficients
        self.as_phase = bool(as_phase)

    def eval_basis(self: OpticalLayer) -> Array:
        """
        Calculates the dot product of the basis vectors and coefficients.

        Returns
        -------
        output : Array
            The output of the dot product between the basis vectors and coefficients.
        """
        return dlu.eval_basis(self.basis, self.coefficients)

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        output = self.eval_basis()
        if self.as_phase:
            wavefront = wavefront.add_phase(output)
        else:
            wavefront += output
        return wavefront


class Tilt(OpticalLayer):
    """
    Tilts the wavefront by the input (x, y) angles.

    ??? abstract "UML"
        ![UML](../../assets/uml/Tilt.png)

    Attributes
    ----------
    angles : Array, radians
        The (x, y) angles by which to tilt the wavefront.
    """

    angles: Array

    def __init__(self: OpticalLayer, angles: Array):
        """
        Parameters
        ----------
        angles : Array, radians
            The (x, y) angles by which to tilt the wavefront.
        """
        super().__init__()
        self.angles = np.asarray(angles, dtype=float)

        if self.angles.shape != (2,):
            raise ValueError("angles must have have (2,)")

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.tilt(self.angles)


class Normalise(OpticalLayer):
    """
    Normalises the wavefront to unit intensity.

    ??? abstract "UML"
        ![UML](../../assets/uml/Normalise.png)
    """

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.normalise()



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import (
    LayeredOpticalSystem,
    AngularOpticalSystem,
    CartesianOpticalSystem,
    PointSource,
    Wavefront,
    PSF,
)
from dLux.layers import Optic


@pytest.fixture
def wf_npixels():
    return 16


@pytest.fixture
def diameter():
    return 1.0


@pytest.fixture
def layers():
    return [Optic()]


def _test_model(optics):
    source = PointSource([1e-6])
    assert isinstance(optics.model(source), np.ndarray)
    assert isinstance(optics.model(source, return_wf=True), Wavefront)
    assert isinstance(optics.model(source, return_psf=True), PSF)
    with pytest.raises(ValueError):
        optics.model(source, return_wf=True, return_psf=True)


def _test_propagate(optics):
    wavels = np.ones(2)
    assert isinstance(optics.propagate(wavels), np.ndarray)
    assert isinstance(optics.propagate(wavels, return_wf=True), Wavefront)
    assert isinstance(optics.propagate(wavels, return_psf=True), PSF)

    with pytest.raises(ValueError):
        optics.propagate(wavels, return_wf=True, return_psf=True)
    with pytest.raises(ValueError):
        optics.propagate(wavels, weights=np.ones(3))
    with pytest.raises(ValueError):
        optics.propagate(wavels, offset=np.ones(3))


def _test_propagate_mono(optics):
    assert isinstance(optics.propagate_mono(1e-6), np.ndarray)
    assert isinstance(optics.propagate_mono(1e-6, return_wf=True), Wavefront)


def test_layered_optics(wf_npixels, diameter, layers):
    optics = LayeredOpticalSystem(wf_npixels, diameter, layers)
    _test_model(optics)
    _test_propagate(optics)
    _test_propagate_mono(optics)

    # Test getattr
    optics.Optic
    optics.opd
    with pytest.raises(AttributeError):
        optics.not_an_attr

    # Test insert and remove layer
    optics.insert_layer(Optic(), 1)
    optics.remove_layer("Optic")


@pytest.fixture
def psf_npixels():
    return 8


@pytest.fixture
def psf_pixel_scale():
    return 1 / 8


@pytest.fixture
def oversample():
    return 2


@pytest.fixture
def focal_length():
    return 1.0


def test_angular_optics(
    wf_npixels, diameter, layers, psf_npixels, psf_pixel_scale, oversample
):
    optics = AngularOpticalSystem(
        wf_npixels, diameter, layers, psf_npixels, psf_pixel_scale, oversample
    )
    _test_model(optics)
    _test_propagate(optics)
    _test_propagate_mono(optics)


def test_cartesian_optics(
    wf_npixels,
    diameter,
    layers,
    focal_length,
    psf_npixels,
    psf_pixel_scale,
    oversample,
):
    optics = CartesianOpticalSystem(
        wf_npixels,
        diameter,
        layers,
        focal_length,
        psf_npixels,
        psf_pixel_scale,
        oversample,
    )
    _test_model(optics)
    _test_propagate(optics)
    _test_propagate_mono(optics)



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import PSF


@pytest.fixture
def psf():
    return PSF(np.ones((16, 16)), 1 / 16)


class TestPSF:
    def test_constructor(self, psf):
        assert psf.npixels == 16
        assert psf.pixel_scale == 1 / 16

    def test_properties(self, psf):
        assert psf.ndim == 0

    def test_methods(self, psf):
        assert psf.downsample(2).npixels == 8
        assert psf.downsample(2).pixel_scale == 1 / 8
        assert isinstance(psf.convolve(np.ones((2, 2))), PSF)
        assert isinstance(psf.convolve(np.ones((2, 2)), method="fft"), PSF)
        assert isinstance(psf.rotate(np.pi), PSF)
        assert isinstance(psf.resize(8), PSF)
        assert isinstance(psf.flip(0), PSF)

    def test_magic(self, psf):
        psf *= np.ones(1)
        assert isinstance(psf, PSF)

        psf += np.ones(1)
        assert isinstance(psf, PSF)

        psf -= np.ones(1)
        assert isinstance(psf, PSF)

        psf /= np.ones(1)
        assert isinstance(psf, PSF)



import jax.numpy as np
from jax import Array

__all__ = [
    "rad2arcsec",
    "rad2deg",
    "rad2arcmin",
    "deg2rad",
    "deg2arcmin",
    "deg2arcsec",
    "arcmin2rad",
    "arcmin2deg",
    "arcmin2arcsec",
    "arcsec2rad",
    "arcsec2deg",
    "arcsec2arcmin",
]


# Radians to:
def rad2arcsec(values: Array) -> Array:
    """
    Converts the inputs values from radians to arcseconds.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 3600 * 180 / np.pi


def rad2deg(values: Array) -> Array:
    """
    Converts the inputs values from radians to degrees.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values * 180 / np.pi


def rad2arcmin(values: Array) -> Array:
    """
    Converts the inputs values from radians to arcminutes.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values * 60 * 180 / np.pi


# Degrees to:
def deg2rad(values: Array) -> Array:
    """
    Converts the inputs values from degrees to radians.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / 180


def deg2arcmin(values: Array) -> Array:
    """
    Converts the inputs values from degrees to arcminutes.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values * 60


def deg2arcsec(values: Array) -> Array:
    """
    Converts the inputs values from degrees to arcseconds.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 3600


# Arcminutes to:
def arcmin2rad(values: Array) -> Array:
    """
    Converts the inputs values from arcminutes to radians.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / (60 * 180)


def arcmin2deg(values: Array) -> Array:
    """
    Converts the inputs values from arcminutes to degrees.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values / 60


def arcmin2arcsec(values: Array) -> Array:
    """
    Converts the inputs values from arcminutes to arcseconds.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into
        arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 60


# Arcseconds to:
def arcsec2rad(values: Array) -> Array:
    """
    Converts the inputs values from arcseconds to radians.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / (3600 * 180)


def arcsec2deg(values: Array) -> Array:
    """
    Converts the inputs values from arcseconds to degrees.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values / 3600


def arcsec2arcmin(values: Array) -> Array:
    """
    Converts the inputs values from arcseconds to arcminutes.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into
        arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values / 60



import jax.numpy as np
import jax.tree as jtu
from jax import Array
from typing import Union

__all__ = [
    "cart2polar",
    "polar2cart",
    "pixel_coords",
    "nd_coords",
    "translate_coords",
    "compress_coords",
    "shear_coords",
    "rotate_coords",
]


def translate_coords(coords: Array, translation: Array) -> Array:
    """
    Translates the coordinates by to a new centre. Translation must have shape (2,).

    Parameters
    ----------
    coords : Array
        The input coordinates to translate.
    translation : Array
        The translation to apply to the coordinates.

    Returns
    -------
    coords : Array
        The translated coordinates.
    """
    return coords - translation[:, None, None]


def compress_coords(coords: Array, compress: Array) -> Array:
    """
    Compresses the coordinates by a given factor. Compress must have shape (2,).

    Parameters
    ----------
    coords : Array
        The input coordinates to compress.
    compress : Array
        The compression to apply to the coordinates.

    Returns
    -------
    coords : Array
        The compressed coordinates.
    """
    return coords * compress[:, None, None]


def shear_coords(coords: Array, shear: Array) -> Array:
    """
    Shears the coordinates by a given factor. Shear must have shape (2,).

    Parameters
    ----------
    coords : Array
        The input coordinates to shear.
    shear : Array
        The shear to apply to the coordinates.

    Returns
    -------
    coords : Array
        The sheared coordinates.
    """
    trans_coords = np.transpose(coords, (0, 2, 1))
    return coords + trans_coords * shear[:, None, None]


def rotate_coords(coords: Array, rotation: float) -> Array:
    """
    Rotates the coordinates by a given angle.

    Parameters
    ----------
    coords : Array
        The input coordinates to rotate.
    rotation : float, radians
        The rotation to apply to the coordinates.

    Returns
    -------
    coords : Array
        The rotated coordinates.
    """

    x, y = coords
    new_x = np.cos(-rotation) * x + np.sin(-rotation) * y
    new_y = -np.sin(-rotation) * x + np.cos(-rotation) * y
    return np.array([new_x, new_y])


# Coordinate conversions #
def cart2polar(coordinates: Array) -> Array:
    """
    Converts the input (x, y) Cartesian coordinates into (r, phi) polar
    coordinates.

    Parameters
    ----------
    coordinates : Array
        The (x, y) Cartesian coordinates to be converted into polar
        coordinates.

    Returns
    -------
    coordinates : Array
        The input Cartesian coordinates converted into (r, phi) polar
        coordinates.
    """
    x, y = coordinates
    return np.array([np.hypot(x, y), np.arctan2(y, x)])


def polar2cart(coordinates: Array) -> Array:
    """
    Converts the input (r, phi) polar coordinates into (x, y) Cartesian
    coordinates.

    Parameters
    ----------
    coordinates : Array
        The (r, phi) polar coordinates to be converted into Cartesian
        coordinates.

    Returns
    -------
    coordinates : Array
        The input polar coordinates converted into (x, y) Cartesian
        coordinates.
    """
    r, phi = coordinates
    return np.array([r * np.cos(phi), r * np.sin(phi)])


# Positions Calculations #
def pixel_coords(npixels: int, diameter: float, polar: bool = False) -> Array:
    """
    Returns a paraxial set of 2d coordinates for each pixel center.

    Parameters
    ----------
    npixels : int
        The output size of the coordinates array to generate.
    diameter : float
        The diameter of the coordinates array to generate.
    polar : bool = False
        Output the coordinates in polar (r, phi) coordinates.

    Returns
    -------
    coordinates : Array
        The array of pixel center coordinates.
    """
    coords = nd_coords((npixels,) * 2, (diameter / npixels,) * 2)
    if polar:
        return cart2polar(coords)
    return coords


def nd_coords(
    npixels: Union[int, tuple],
    pixel_scales: Union[tuple, float] = 1.0,
    offsets: Union[tuple, float] = 0.0,
    indexing: str = "xy",
) -> Array:
    """
    Returns a set of nd pixel center coordinates, with an optional offset. Each
    dimension can have a different number of pixels, pixel scale and offset by passing
    in tuples of values: `nd_coords((10, 10), (1, 2), (0, 1))`. pixel scale and offset
    can also be passed in as floats to apply those values to all dimensions, ie:
    `nd_coords((10, 10), 1, 0)`.

    The indexing argument is the same as in numpy.meshgrid., i.e.: Giving the
    string ‘ij’ returns a meshgrid with matrix indexing, while ‘xy’ returns a
    meshgrid with Cartesian indexing. In the 2-D case with inputs of length M
    and N, the outputs are of shape (N, M) for ‘xy’ indexing and (M, N) for
    ‘ij’ indexing. In the 3-D case with inputs of length M, N and P, outputs
    are of shape (N, M, P) for ‘xy’ indexing and (M, N, P) for ‘ij’ indexing.

    Parameters
    ----------
    npixels : Union[int, tuple]
        The number of pixels in each dimension.
    pixel_scales : Union[tuple, float] = 1.
        The pixel_scales of each dimension. If a tuple, the length
        of the tuple must match the number of dimensions. If a float, the same
        scale is applied to all dimensions.
    offsets : Union[tuple, float] = 0.
        The offset of the pixel centers in each dimension. If a tuple, the
        length of the tuple must match the number of dimensions. If a float,
        the same offset is applied to all dimensions.
        set to 0.
    indexing : str = 'xy'
        The indexing of the output. Default is 'xy'. See numpy.meshgrid for
        more details.

    Returns
    -------
    coordinates : Array
        The positions of the pixel centers in the given dimensions.
    """
    if indexing not in ["xy", "ij"]:
        raise ValueError("indexing must be either 'xy' or 'ij'.")

    # Promote npixels to tuple to handle 1d case
    if not isinstance(npixels, tuple):
        npixels = (npixels,)

    # Assume equal pixel scales if not given
    if not isinstance(pixel_scales, tuple):
        pixel_scales = (pixel_scales,) * len(npixels)

    # Assume no offset if not given
    if not isinstance(offsets, tuple):
        offsets = (offsets,) * len(npixels)

    def pixel_fn(n, offset, scale):
        start = -(n - 1) / 2 * scale - offset
        end = (n - 1) / 2 * scale - offset
        return np.linspace(start, end, n)

    # Generate the linear edges of each axes
    # TODO: tree.flatten()[0] to avoid squeeze?
    lin_pixels = jtu.map(pixel_fn, npixels, offsets, pixel_scales)

    # output (x, y) for 2d, else in order
    positions = np.array(np.meshgrid(*lin_pixels, indexing=indexing))

    # Squeeze for empty axis removal in 1d case
    return np.squeeze(positions)



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import Wavefront, Optic


@pytest.fixture
def wavefront():
    return Wavefront(npixels=16, diameter=1.0, wavelength=1e-6)


# TODO: Test magic
class TestWavefront:
    def test_constructor(self, wavefront):
        assert wavefront.npixels == 16
        assert wavefront.diameter == 1.0
        assert wavefront.wavelength == 1e-6

    def test_properties(self, wavefront):
        assert wavefront.ndim == 0
        assert wavefront.real.shape == wavefront.amplitude.shape
        assert wavefront.imaginary.shape == wavefront.amplitude.shape
        assert wavefront.phasor.shape == wavefront.amplitude.shape
        assert wavefront.psf.shape == wavefront.amplitude.shape
        assert wavefront.coordinates.shape == (2, 16, 16)
        assert np.array(wavefront.wavenumber).shape == ()
        assert np.array(wavefront.fringe_size).shape == ()

    def test_methods(self, wavefront):
        assert isinstance(wavefront.add_opd(0), Wavefront)
        assert isinstance(wavefront.add_phase(0), Wavefront)
        assert isinstance(wavefront.tilt(np.zeros(2)), Wavefront)
        with pytest.raises(ValueError):
            wavefront.tilt(np.zeros(3))
        assert isinstance(wavefront.normalise(), Wavefront)
        assert isinstance(wavefront.flip(0), Wavefront)
        assert wavefront.scale_to(8, 1 / 32).npixels == 8
        assert np.allclose(wavefront.scale_to(8, 1 / 32).pixel_scale, 1 / 32)
        assert np.allclose(
            wavefront.scale_to(8, 1 / 32, True).pixel_scale, 1 / 32
        )
        assert isinstance(wavefront.rotate(np.pi), Wavefront)
        assert isinstance(wavefront.resize(8), Wavefront)

    def test_magic_add(self, wavefront):
        # Test Nones
        wavefront += None
        assert isinstance(wavefront, Wavefront)

        # Test optical layers
        wavefront += Optic()
        assert isinstance(wavefront, Wavefront)

        # Test arrays
        wavefront += np.ones(1)
        assert isinstance(wavefront, Wavefront)

        # Test numeric
        wavefront += 1
        assert isinstance(wavefront, Wavefront)

        # Test invalid
        with pytest.raises(TypeError):
            wavefront += "1"

    def test_magic_mul(self, wavefront):
        # Test Nones
        wavefront *= None
        assert isinstance(wavefront, Wavefront)

        # Test optical layers
        wavefront *= Optic()
        assert isinstance(wavefront, Wavefront)

        # Test arrays
        wavefront *= np.ones(1)
        assert isinstance(wavefront, Wavefront)

        # Test complex arrays
        wavefront *= np.ones(1) * np.exp(1j)
        assert isinstance(wavefront, Wavefront)

        # Test numeric
        wavefront *= 1
        assert isinstance(wavefront, Wavefront)

        # Test invalid
        with pytest.raises(TypeError):
            wavefront *= "1"

    def _test_propagated(self, wf, plane, units, npix, pscale):
        assert wf.plane == plane
        assert wf.units == units
        assert wf.npixels == npix
        assert np.allclose(wf.pixel_scale, pscale)

    @pytest.mark.parametrize("pixel", [True, False])
    def test_propagation(self, wavefront, pixel):
        npix = 8
        pscale = 1 / 32
        focal_length = 1.0

        # No focal length to focal plane
        focal_wf = wavefront.propagate(npix, pscale, pixel=pixel)
        self._test_propagated(focal_wf, "Focal", "Angular", npix, pscale)

        # No focal length to pupil plane
        pupil_wf = focal_wf.propagate(npix, pscale, pixel=pixel)
        self._test_propagated(pupil_wf, "Pupil", "Cartesian", npix, pscale)

        # From an angular focal plane, specifying fl
        with pytest.raises(ValueError):
            focal_wf.propagate(npix, pscale, focal_length, pixel=pixel)

        # Focal length to focal plane
        focal_wf = wavefront.propagate(npix, pscale, focal_length, pixel=pixel)
        self._test_propagated(focal_wf, "Focal", "Cartesian", npix, pscale)

        # Focal length to pupil plane
        pupil_wf = focal_wf.propagate(npix, pscale, focal_length, pixel=pixel)
        self._test_propagated(pupil_wf, "Pupil", "Cartesian", npix, pscale)

    def test_fft_propagation(self, wavefront):
        pad = 2
        focal_length = 1.0
        npix_out = wavefront.npixels * pad
        pscale_in = wavefront.pixel_scale
        pscale_out = wavefront.fringe_size / pad

        # No focal length to focal plane
        focal_wf = wavefront.propagate_FFT()
        self._test_propagated(
            focal_wf, "Focal", "Angular", npix_out, pscale_out
        )

        # No focal length to pupil plane
        pupil_wf = focal_wf.resize(wavefront.npixels).propagate_FFT()
        self._test_propagated(
            pupil_wf, "Pupil", "Cartesian", npix_out, pscale_in
        )

        # From an angular focal plane, specifying fl
        with pytest.raises(ValueError):
            focal_wf.propagate_FFT(focal_length)

        # With focal length
        pscale_in *= focal_length

        # Focal length to focal plane
        focal_wf = wavefront.propagate_FFT(focal_length)
        self._test_propagated(
            focal_wf, "Focal", "Cartesian", npix_out, pscale_out
        )

        # Focal length to pupil plane
        pupil_wf = focal_wf.resize(wavefront.npixels).propagate_FFT(
            focal_length
        )
        self._test_propagated(
            pupil_wf, "Pupil", "Cartesian", npix_out, pscale_in
        )

    @pytest.mark.parametrize("pixel", [True, False])
    def test_fresnel_propagation(self, wavefront, pixel):
        npix = 8
        pscale = 1 / 32
        focal_length = 1.0
        focal_shift = 1e-2

        # Simple test
        focal_wf = wavefront.propagate_fresnel(
            npix, pscale, focal_length, focal_shift, pixel=pixel
        )
        self._test_propagated(
            focal_wf, "Intermediate", "Cartesian", npix, pscale
        )

        # Test error from focal plane
        focal_wf = wavefront.propagate(npix, pscale, pixel=pixel)
        with pytest.raises(ValueError):
            focal_wf.propagate_fresnel(
                npix, pscale, focal_length, focal_shift, pixel=pixel
            )



import jax.numpy as np
from jax import Array
from jax.scipy.ndimage import map_coordinates
import dLux.utils as dlu

__all__ = [
    "generate_coordinates",
    "scale",
    "rotate",
]


def generate_coordinates(
    npixels_in: int,
    npixels_out: int,
    sampling_ratio: Array,
    x_shift: Array = np.array(0.0),
    y_shift: Array = np.array(0.0),
) -> Array:
    """
    Generates a new set of paraxial coordinates which can be used for interpolation.

    Parameters
    ----------
    npixels_in : int
        The number of pixels in the original array.
    npixels_out : int
        The number of pixel in the output array.
    sampling_ratio : Array
        The ratio of pixel sizes in the input and output array,
        i.e. pixel_scale_out/pixel_scale_in.
    x_shift : Array, pixels = np.array(0.)
        How much to shift the x_coordinates in the output array, in the pixel
        units of the output array.
    y_shift : Array, pixels = np.array(0.)
        How much to shift the y_coordinates in the output array, in the pixel
        units of the output array.

    Returns
    -------
    coordinates : Array
        The output coordinates at which to interpolate onto.
    """
    old_centre = (npixels_in - 1) / 2
    new_centre = (npixels_out - 1) / 2
    pixels = (
        sampling_ratio * np.linspace(-new_centre, new_centre, npixels_out)
        + old_centre
    )
    x_pixels, y_pixels = np.meshgrid(pixels + x_shift, pixels + y_shift)
    return np.array([y_pixels, x_pixels])


def scale(array: Array, npixels: int, ratio: float, order: int = 1) -> Array:
    """
    Paraxially interpolates an array based on the sampling ratio, and npixels_out.

    # TODO: Check if a half-pixel offset is produced

    Parameters
    ----------
    array : Array
        The input field to interpolate, either in amplitude and phase, or real
        and imaginary.
    npixels : int
        The number of pixel in the output array.
    ratio : float
        The relative input to output scales, TODO: does 2 make it bigger or
        smaller? i.e. input scale/output scale. <- get this right.
    order : int = 1
        The interpolation order to use. Can be 0 or 1.

    Returns
    -------
    array : Array
        The interpolated array.
    """
    # Get coords arrays
    npixels_in = array.shape[-1]
    # TODO: Update with array_coordinates
    coordinates = generate_coordinates(npixels_in, npixels, ratio)
    return map_coordinates(array, coordinates, order=order)


def rotate(array: Array, angle: Array, order: int = 1) -> Array:
    """
    Rotates an array by the angle, using interpolation.

    Parameters
    ----------
    array : Array
        The array to rotate.
    angle : Array, radians
        The angle to rotate the array by.
    order : int = 1
        The interpolation order to use. Can be 0 or 1.

    Returns
    -------
    array : Array
        The rotated array.
    """

    # TODO: Use rotate_coords
    def _rotate(coordinates: Array, rotation: Array) -> Array:
        x, y = coordinates[0], coordinates[1]
        new_x = np.cos(-rotation) * x + np.sin(-rotation) * y
        new_y = -np.sin(-rotation) * x + np.cos(-rotation) * y
        return np.array([new_x, new_y])

    # Get coordinates
    npixels = array.shape[0]
    centre = (npixels - 1) / 2
    coordinates = dlu.nd_coords((npixels, npixels), indexing="ij")
    coordinates_rotated = _rotate(coordinates, angle) + centre

    # Interpolate
    return map_coordinates(array, coordinates_rotated, order=order)



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import CoordTransform


def test_coord_transform():
    CoordTransform().calculate(1, 16)
    CoordTransform([0.0, 0.0], np.pi, [1, 1], [1, 1]).calculate(1, 16)
    with pytest.raises(ValueError):
        CoordTransform(translation=[0.0])
    with pytest.raises(ValueError):
        CoordTransform(rotation=[0.0])
    with pytest.raises(ValueError):
        CoordTransform(compression=[0.0])
    with pytest.raises(ValueError):
        CoordTransform(shear=[0.0])



from __future__ import annotations
from collections import OrderedDict
from abc import abstractmethod
import jax.numpy as np
from jax import Array
from zodiax import filter_vmap, Base
from typing import Union, Any
import dLux.utils as dlu


__all__ = [
    "BaseOpticalSystem",
    "AngularOpticalSystem",
    "CartesianOpticalSystem",
    "LayeredOpticalSystem",
]

from .layers.optical_layers import OpticalLayer
from .wavefronts import Wavefront
from .sources import BaseSource as Source
from .psfs import PSF


###################
# Private Classes #
###################
class BaseOpticalSystem(Base):
    @abstractmethod
    def propagate_mono(
        self: BaseOpticalSystem,
        wavelength: float,
        offset: Array = np.zeros(2),
        return_wf: bool = False,
    ) -> Array:  # pragma: no cover
        """
        Propagates a monochromatic point source through the optical layers.

        Parameters
        ----------
        wavelength : float, metres
            The wavelength of the wavefront to propagate through the optical layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        return_wf: bool = False
            Should the Wavefront object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront
            if `return_wf` is False, returns the psf Array.
            if `return_wf` is True, returns the Wavefront object.
        """

    @abstractmethod
    def propagate(
        self: OpticalSystem,
        wavelengths: Array,
        offset: Array = np.zeros(2),
        weights: Array = None,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:  # pragma: no cover
        """
        Propagates a Polychromatic point source through the optics.

        Parameters
        ----------
        wavelengths : Array, metres
            The wavelengths of the wavefronts to propagate through the optics.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        weights : Array = None
            The weight of each wavelength. If None, all weights are equal.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.

        """

    @abstractmethod
    def model(
        self: OpticalSystem,
        source: Source,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:  # pragma: no cover
        """
        Models the input Source object through the optics.

        Parameters
        ----------
        source : Source
            The Source object to model through the optics.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """


class OpticalSystem(BaseOpticalSystem):
    """
    Base optics class implementing both the `propagate` and `model` methods that are
    universal to all optics classes.
    """

    def propagate(
        self: OpticalSystem,
        wavelengths: Array,
        offset: Array = np.zeros(2),
        weights: Array = None,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Propagates a Polychromatic point source through the optics.

        Parameters
        ----------
        wavelengths : Array, metres
            The wavelengths of the wavefronts to propagate through the optics.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        weights : Array = None
            The weight of each wavelength. If None, all weights are equal.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        if return_wf and return_psf:
            raise ValueError(
                "return_wf and return_psf cannot both be True. "
                "Please choose one."
            )

        wavelengths = np.atleast_1d(wavelengths)
        if weights is None:
            weights = np.ones_like(wavelengths) / len(wavelengths)
        else:
            weights = np.atleast_1d(weights)

        # Check wavelengths and weights
        if weights.shape != wavelengths.shape:
            raise ValueError(
                "wavelengths and weights must have the "
                f"same shape, got {wavelengths.shape} and {weights.shape} "
                "respectively."
            )

        # Check offset
        offset = np.array(offset) if not isinstance(offset, Array) else offset
        if offset.shape != (2,):
            raise ValueError(
                "offset must be a 2-element array, got "
                f"shape {offset.shape}."
            )

        # Calculate - note we multiply by sqrt(weight) to account for the
        # fact that the PSF is the square of the amplitude
        prop_fn = lambda wavelength, weight: self.propagate_mono(
            wavelength, offset, return_wf=True
        ).multiply("amplitude", weight**0.5)
        wf = filter_vmap(prop_fn)(wavelengths, weights)

        # Return PSF, Wavefront, or array psf
        if return_wf:
            return wf
        if return_psf:
            return PSF(wf.psf.sum(0), wf.pixel_scale.mean())
        return wf.psf.sum(0)

    def model(
        self: OpticalSystem,
        source: Source,
        return_wf: bool = False,
        return_psf: bool = False,
    ) -> Array:
        """
        Models the input Source object through the optics.

        Parameters
        ----------
        source : Source
            The Source object to model through the optics.
        return_wf : bool = False
            Should the Wavefront object be returned instead of the psf Array?
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront, PSF
            if `return_wf` is False and `return_psf` is False, returns the psf Array.
            if `return_wf` is True and `return_psf` is False, returns the Wavefront
                object.
            if `return_wf` is False and `return_psf` is True, returns the PSF object.
        """
        return source.model(self, return_wf, return_psf)


class ParametricOpticalSystem(OpticalSystem):
    """
    Implements the attributes required for an optical system with a specific output
    pixel scale and number of pixels.

    Attributes
    ----------
    psf_npixels : int
        The number of pixels of the final PSF.
    oversample : int
        The oversampling factor of the final PSF. Decreases the psf_pixel_scale
        parameter while increasing the psf_npixels parameter.
    psf_pixel_scale : float
        The pixel scale of the final PSF.
    """

    psf_npixels: int
    oversample: int
    psf_pixel_scale: float

    def __init__(
        self: OpticalSystem,
        psf_npixels: int,
        psf_pixel_scale: float,
        oversample: int = 1,
        **kwargs,
    ):
        """
        Parameters
        ----------
        psf_npixels : int
            The number of pixels of the final PSF.
        psf_pixel_scale : float
            The pixel scale of the final PSF.
        oversample : int = 1.
            The oversampling factor of the final PSF. Decreases the psf_pixel_scale
            parameter while increasing the psf_npixels parameter.
        """
        self.psf_npixels = int(psf_npixels)
        self.oversample = int(oversample)
        self.psf_pixel_scale = float(psf_pixel_scale)
        super().__init__(**kwargs)


##################
# Public Classes #
##################
class LayeredOpticalSystem(OpticalSystem):
    """
    A flexible optical system that allows for the arbitrary chaining of OpticalLayers.

    ??? abstract "UML"
        ![UML](../../assets/uml/LayeredOpticalSystem.png)

    Attributes
    ----------
    wf_npixels : int
        The size of the initial wavefront to propagate.
    diameter : float, metres
        The diameter of the wavefront to propagate.
    layers : OrderedDict
        A series of `OpticalLayer` transformations to apply to wavefronts.
    """

    wf_npixels: int
    diameter: float
    layers: OrderedDict

    def __init__(
        self: OpticalSystem,
        wf_npixels: int,
        diameter: float,
        layers: list[OpticalLayer, tuple],
    ):
        """
        Parameters
        ----------
        wf_npixels : int
            The size of the initial wavefront to propagate.
        diameter : float
            The diameter of the wavefront to propagate.
        layers : list[OpticalLayer, tuple]
            A list of `OpticalLayer` transformations to apply to wavefronts. The list
            entries can be either `OpticalLayer` objects or tuples of (key, layer) to
            specify a key for the layer in the layers dictionary.
        """
        self.wf_npixels = int(wf_npixels)
        self.diameter = float(diameter)
        self.layers = dlu.list2dictionary(layers, True, OpticalLayer)

    def __getattr__(self: OpticalSystem, key: str) -> Any:
        """
        Raises both the individual layers and the attributes of the layers via
        their keys.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the layers dictionary.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the layers dictionary.
        """
        if key in self.layers.keys():
            return self.layers[key]
        for layer in list(self.layers.values()):
            if hasattr(layer, key):
                return getattr(layer, key)
        raise AttributeError(
            f"{self.__class__.__name__} has no attribute " f"{key}."
        )

    def propagate_mono(
        self: OpticalSystem,
        wavelength: Array,
        offset: Array = np.zeros(2),
        return_wf: bool = False,
    ) -> Array:
        """
        Propagates a monochromatic point source through the optical layers.

        Parameters
        ----------
        wavelength : float, metres
            The wavelength of the wavefront to propagate through the optical layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        return_wf: bool = False
            Should the Wavefront object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront
            if `return_wf` is False, returns the psf Array.
            if `return_wf` is True, returns the Wavefront object.
        """
        # Initialise wavefront
        wavefront = Wavefront(self.wf_npixels, self.diameter, wavelength)
        wavefront = wavefront.tilt(offset)

        # Apply layers
        for layer in list(self.layers.values()):
            wavefront *= layer

        # Return PSF or Wavefront
        if return_wf:
            return wavefront
        return wavefront.psf

    def insert_layer(
        self: OpticalSystem, layer: Union[OpticalLayer, tuple], index: int
    ) -> OpticalSystem:
        """
        Inserts a layer into the layers dictionary at a specified index. This function
        calls the list2dictionary function to ensure all keys remain unique. Note that
        this can result in some keys being modified if they are duplicates. The input
        'layer' can be a tuple of (key, layer) to specify a key, else the key is taken
        as the class name of the layer.

        Parameters
        ----------
        layer : Any
            The layer to be inserted.
        index : int
            The index at which to insert the layer.

        Returns
        -------
        optical_system : OpticalSystem
            The updated optical system.
        """
        return self.set(
            "layers", dlu.insert_layer(self.layers, layer, index, OpticalLayer)
        )

    def remove_layer(self: OpticalLayer, key: str) -> OpticalSystem:
        """
        Removes a layer from the layers dictionary, specified by its key.

        Parameters
        ----------
        key : str
            The key of the layer to be removed.

        Returns
        -------
        optical_system : OpticalSystem
            The updated optical system.
        """
        return self.set("layers", dlu.remove_layer(self.layers, key))


class AngularOpticalSystem(ParametricOpticalSystem, LayeredOpticalSystem):
    """
    An extension to the LayeredOpticalSystem class that propagates a wavefront to an
    image plane with `psf_pixel_scale` in units of arcseconds.

    ??? abstract "UML"
        ![UML](../../assets/uml/AngularOpticalSystem.png)

    Attributes
    ----------
    wf_npixels : int
        The number of pixels representing the wavefront.
    diameter : Array, metres
        The diameter of the initial wavefront to propagate.
    layers : OrderedDict
        A series of `OpticalLayer` transformations to apply to wavefronts.
    psf_npixels : int
        The number of pixels of the final PSF.
    psf_pixel_scale : float, arcseconds
        The pixel scale of the final PSF.
    oversample : int
        The oversampling factor of the final PSF. Decreases the psf_pixel_scale
        parameter while increasing the psf_npixels parameter.
    """

    def __init__(
        self: OpticalSystem,
        wf_npixels: int,
        diameter: float,
        layers: list[OpticalLayer, tuple],
        psf_npixels: int,
        psf_pixel_scale: float,
        oversample: int = 1,
    ):
        """
        Parameters
        ----------
        wf_npixels : int
            The number of pixels representing the wavefront.
        diameter : Array, metres
            The diameter of the initial wavefront to propagate.
        layers : list[OpticalLayer, tuple]
            A list of `OpticalLayer` transformations to apply to wavefronts. The list
            entries can be either `OpticalLayer` objects or tuples of (key, layer) to
            specify a key for the layer in the layers dictionary.
        psf_npixels : int
            The number of pixels of the final PSF.
        psf_pixel_scale : float, arcseconds
            The pixel scale of the final PSF in units of arcseconds.
        oversample : int
            The oversampling factor of the final PSF. Decreases the psf_pixel_scale
            parameter while increasing the psf_npixels parameter.
        """
        super().__init__(
            wf_npixels=wf_npixels,
            diameter=diameter,
            layers=layers,
            psf_npixels=psf_npixels,
            psf_pixel_scale=psf_pixel_scale,
            oversample=oversample,
        )

    def propagate_mono(
        self: OpticalSystem,
        wavelength: Array,
        offset: Array = np.zeros(2),
        return_wf: bool = False,
    ) -> Array:
        """
        Propagates a monochromatic point source through the optical layers.

        Parameters
        ----------
        wavelength : float, metres
            The wavelength of the wavefront to propagate through the optical layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        return_wf: bool = False
            Should the Wavefront object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront
            if `return_wf` is False, returns the psf Array.
            if `return_wf` is True, returns the Wavefront object.
        """
        wf = super().propagate_mono(wavelength, offset, return_wf=True)

        # Propagate
        true_pixel_scale = self.psf_pixel_scale / self.oversample
        pixel_scale = dlu.arcsec2rad(true_pixel_scale)
        psf_npixels = self.psf_npixels * self.oversample
        wf = wf.propagate(psf_npixels, pixel_scale)

        # Return PSF or Wavefront
        if return_wf:
            return wf
        return wf.psf


class CartesianOpticalSystem(ParametricOpticalSystem, LayeredOpticalSystem):
    """
    An extension to the LayeredOpticalSystem class that propagates a wavefront to an
    image plane with `psf_pixel_scale` in units of microns.

    ??? abstract "UML"
        ![UML](../../assets/uml/CartesianOpticalSystem.png)

    Attributes
    ----------
    wf_npixels : int
        The number of pixels representing the wavefront.
    diameter : Array, metres
        The diameter of the initial wavefront to propagate.
    layers : OrderedDict
        A series of `OpticalLayer` transformations to apply to wavefronts.
    focal_length : float, metres
        The focal length of the system.
    psf_npixels : int
        The number of pixels of the final PSF.
    psf_pixel_scale : float, microns
        The pixel scale of the final PSF.
    oversample : int
        The oversampling factor of the final PSF. Decreases the psf_pixel_scale
        parameter while increasing the psf_npixels parameter.
    """

    focal_length: None

    def __init__(
        self: OpticalSystem,
        wf_npixels: int,
        diameter: float,
        layers: list[OpticalLayer, tuple],
        focal_length: float,
        psf_npixels: int,
        psf_pixel_scale: float,
        oversample: int = 1,
    ):
        """
        Parameters
        ----------
        wf_npixels : int
            The number of pixels representing the wavefront.
        diameter : Array, metres
            The diameter of the initial wavefront to propagate.
        layers : list[OpticalLayer, tuple]
            A list of `OpticalLayer` transformations to apply to wavefronts. The list
            entries can be either `OpticalLayer` objects or tuples of (key, layer) to
            specify a key for the layer in the layers dictionary.
        focal_length : float, metres
            The focal length of the system.
        psf_npixels : int
            The number of pixels of the final PSF.
        psf_pixel_scale : float, microns
            The pixel scale of the final PSF in units of microns.
        oversample : int
            The oversampling factor of the final PSF. Decreases the psf_pixel_scale
            parameter while increasing the psf_npixels parameter.
        """
        self.focal_length = float(focal_length)

        super().__init__(
            wf_npixels=wf_npixels,
            diameter=diameter,
            layers=layers,
            psf_npixels=psf_npixels,
            psf_pixel_scale=psf_pixel_scale,
            oversample=oversample,
        )

    def propagate_mono(
        self: OpticalSystem,
        wavelength: Array,
        offset: Array = np.zeros(2),
        return_wf: bool = False,
    ) -> Array:
        """
        Propagates a monochromatic point source through the optical layers.

        Parameters
        ----------
        wavelength : float, metres
            The wavelength of the wavefront to propagate through the optical layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        return_wf: bool = False
            Should the Wavefront object be returned instead of the psf Array?

        Returns
        -------
        object : Array, Wavefront
            if `return_wf` is False, returns the psf Array.
            if `return_wf` is True, returns the Wavefront object.
        """
        wf = super().propagate_mono(wavelength, offset, return_wf=True)

        # Propagate
        true_pixel_scale = self.psf_pixel_scale / self.oversample
        pixel_scale = 1e-6 * true_pixel_scale
        psf_npixels = self.psf_npixels * self.oversample
        wf = wf.propagate(psf_npixels, pixel_scale, self.focal_length)

        # Return PSF or Wavefront
        if return_wf:
            return wf
        return wf.psf



import jax.numpy as np
from jax import lax, vmap, Array
import dLux.utils as dlu

__all__ = [
    "combine",
    "circle",
    "square",
    "rectangle",
    "reg_polygon",
    "spider",
    "soft_circle",
    "soft_square",
    "soft_rectangle",
    "soft_reg_polygon",
    "soft_spider",
]


def combine(arrays: Array, oversample: int = 1, sum: bool = False) -> Array:
    """
    Combines multiple arrays by multiplying them together, and downsampling the output.

    Parameters
    ----------
    arrays : Array
        The arrays to be combined. Should have shape (n_arrays, npix, npix).
    oversample : int = 1
        The amount to downsample the output by.
    sum : bool = False
        Whether to sum the arrays instead of multiplying them.
    """
    method = np.sum if sum else np.prod
    array = np.array(arrays)
    if oversample == 1:
        return method(array, 0)
    return dlu.downsample(method(array, 0), oversample)


def shift_and_scale(array: Array) -> Array:
    """
    Shifts and scales the array to be between 0 and 1

    Parameters
    ----------
    array : Array
        The array to be shifted and scaled.

    Returns
    -------
    array : Array
        The shifted and scaled array.
    """
    return dlu.math.nandiv(
        array - array.min(), array.max() - array.min(), fill=0
    )


def soften(distances: Array, clip_dist: float, invert: bool = False) -> Array:
    """
    Softens the edges of a distances array by clipping the distances to a maximum
    value, and then shifting and scaling the array to be between 0 and 1.

    Parameters
    ----------
    distances : Array
        The array of distances to be softened.
    clip_dist : float
        The maximum distance to clip the distances to.
    invert : bool = False
        Whether to invert the distances before softening.

    Returns
    -------
    distances : Array
        The softened distances array.
    """
    # TODO: Possibly clip from -clip_dist:0 to ensure zernikes have full pupil support
    if invert:
        distances *= -1
    return shift_and_scale(np.clip(distances, -clip_dist, clip_dist))


#####################
def circle(coords: Array, radius: float, invert: bool = False) -> Array:
    """
    Calculates a soft-edged circle via downsampling. This function is not
    differentiable.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the circle on.
    radius : float
        The radius of the circle.
    invert : bool = False
        Whether to invert the circle.

    Returns
    -------
    circle : Array
        The circle.
    """
    if invert:
        return (circ_distance(coords, radius) > 0).astype(float)
    return (circ_distance(coords, radius) < 0).astype(float)


def square(coords: Array, width: float, invert: bool = False) -> Array:
    """
    Calculates a soft-edged square via downsampling. This function is not
    differentiable.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the square on.
    width : float
        The width of the square.
    invert : bool = False
        Whether to invert the square.

    Returns
    -------
    square : Array
        The square.
    """
    if invert:
        return (square_distance(coords, width) > 0).astype(float)
    return (square_distance(coords, width) < 0).astype(float)


def rectangle(
    coords: Array, width: float, height: float, invert: bool = False
) -> Array:
    """
    Calculates a soft-edged rectangle via downsampling. This function is not
    differentiable.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the rectangle on.
    width : float
        The width of the rectangle.
    height : float
        The height of the rectangle.

    Returns
    -------
    rectangle : Array
        The rectangle.
    """
    if invert:
        return (rectangle_distance(coords, width, height) > 0).astype(float)
    return (rectangle_distance(coords, width, height) < 0).astype(float)


def reg_polygon(
    coords: Array, rmax: float, nsides: int, invert: bool = False
) -> Array:
    """
    Calculates a soft-edged regular polygon via downsampling. This function is not
    differentiable.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the polygon on.
    rmax : float
        The radius of the polygon.
    nsides : int
        The number of sides of the polygon.
    invert : bool = False
        Whether to invert the polygon.

    Returns
    -------
    polygon : Array
        The polygon.
    """
    if invert:
        return (reg_polygon_distance(coords, nsides, rmax) > 0).astype(float)
    return (reg_polygon_distance(coords, nsides, rmax) < 0).astype(float)


def spider(coords: Array, width: float, angles: Array) -> Array:
    """
    Calculates a soft-edged spider via downsampling. This function is not
    differentiable.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the spider on.
    width : float
        The width of the spider.
    angles : Array
        The angles of the spider.

    Returns
    -------
    spider : Array
        The spider.
    """
    angles = np.array(angles) if not isinstance(angles, np.ndarray) else angles
    calc_fn = vmap(lambda angle: spider_distance(coords, width, angle) < 0)
    return (
        ~lax.reduce(calc_fn(angles), np.array(False), lax.bitwise_or, (0,))
    ).astype(float)


# TODO: This eventually
# def irreg_polygon(
# npix, diam, vertices, oversample=1, invert=False, shift=np.zeros(2)):
#     """Calcs an downsampled irregular polygon to gain soft edges"""
#     pass


################
### Softened ###
################
def soft_circle(
    coords: Array, radius: float, clip_dist: float = 0.1, invert: bool = False
) -> Array:
    """
    Calculates a soft-edged circle differentiably. The 'clip_dist' parameter defines
    the distance from the edge to 'soften' up to. A large clip_dist will result in a
    circle with a very soft edge, while a small clip_dist will result in a circle with
    a very hard edge.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the circle on.
    radius : float
        The radius of the circle.
    clip_dist : float = 0.1
        The distance from the edge to 'soften' up to.
        circle.
    invert : bool = False
        Whether to invert the circle.

    Returns
    -------
    circle : Array
        The softened circle.
    """
    distances = -circ_distance(coords, radius)
    return soften(distances, clip_dist, invert)


def soft_square(
    coords: Array, width: float, clip_dist: float = 0.1, invert: bool = False
) -> Array:
    """
    Calculates a soft-edged square differentiably. The 'clip_dist' parameter defines
    the distance from the edge to 'soften' up to. A large clip_dist will result in a
    square with a very soft edge, while a small clip_dist will result in a square with
    a very hard edge.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the square on.
    width : float
        The width of the square.
    clip_dist : float = 0.1
        The distance from the edge to 'soften' up to.
        square.
    invert : bool = False
        Whether to invert the square.

    Returns
    -------
    square : Array
        The softened square.
    """
    distances = -square_distance(coords, width)
    return soften(distances, clip_dist, invert)


def soft_rectangle(
    coords: Array,
    width: float,
    height: float,
    clip_dist: float = 0.1,
    invert: bool = False,
) -> Array:
    """
    Calculates a soft-edged rectangle differentiably. The 'clip_dist' parameter defines
    the distance from the edge to 'soften' up to. A large clip_dist will result in a
    rectangle with a very soft edge, while a small clip_dist will result in a rectangle
    with a very hard edge.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the rectangle on.
    width : float
        The width of the rectangle.
    height : float
        The height of the rectangle.
    clip_dist : float = 0.1
        The distance from the edge to 'soften' up to.

    Returns
    -------
    rectangle : Array
        The softened rectangle.
    """
    distances = -rectangle_distance(coords, width, height)
    return soften(distances, clip_dist, invert)


def soft_reg_polygon(
    coords: Array,
    radius: float,
    nsides: int,
    clip_dist: float = 0.1,
    invert: bool = False,
) -> Array:
    """
    Calculates a soft-edged regular polygon differentiably. The 'clip_dist' parameter
    defines the distance from the edge to 'soften' up to. A large clip_dist will result
    in a polygon with a very soft edge, while a small clip_dist will result in a polygon
    with a very hard edge.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the polygon on.
    radius : float
        The radius of the polygon.
    nsides : int
        The number of sides of the polygon.
    clip_dist : float = 0.1
        The distance from the edge to 'soften' up to.
    invert : bool = False
        Whether to invert the polygon.

    Returns
    -------
    polygon : Array
        The softened polygon.
    """
    distances = -reg_polygon_distance(coords, nsides, radius)
    return soften(distances, clip_dist, invert)


def soft_spider(
    coords: Array,
    width: float,
    angles: Array,
    clip_dist: float = 0.1,
    invert: bool = False,
) -> Array:
    """
    Calculates a soft-edged spider differentiably. The 'clip_dist' parameter defines
    the distance from the edge to 'soften' up to. A large clip_dist will result in a
    spider with a very soft edge, while a small clip_dist will result in a spider with
    a very hard edge.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the spider on.
    width : float
        The width of the spider.
    angles : Array
        The angles of the spider.
    clip_dist : float = 0.1
        The distance from the edge to 'soften' up to.
    invert : bool = False
        Whether to invert the spider.

    Returns
    -------
    spider : Array
        The softened spider.
    """
    angles = np.array(angles) if not isinstance(angles, np.ndarray) else angles
    spider_fn = vmap(lambda angle: spider_distance(coords, width, angle))
    spiders = -spider_fn(angles).min(axis=0)
    return soften(spiders, clip_dist, invert)


# def soft_irreg_polygon(radius, coords, clip_dist=0.1, invert=False):
#     """Dynamically calculates a soft irregular polygon differentiably"""
#     pass


##########################
### Distance functions ###
##########################
def circ_distance(coords: Array, radius: float) -> Array:
    """
    Calculates the distance from the edge of a circle.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the circle.
    radius : float
        The radius of the circle.

    Returns
    -------
    distance : Array
        The distance from the edge of the circle.
    """
    return dlu.cart2polar(coords)[0] - radius


def square_distance(coords: Array, width: float) -> Array:
    """
    Calculates the distance from the edge of a square.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the square.
    width : float
        The width of the square.

    Returns
    -------
    distance : Array
        The distance from the edge of the square.
    """
    return np.max(np.abs(coords), axis=0) - width / 2


def rectangle_distance(coords: Array, width: float, height: float) -> Array:
    """
    Calculates the distance from the edge of a rectangle.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the rectangle.
    width : float
        The width of the rectangle.
    height : float
        The height of the rectangle.

    Returns
    -------
    distance : Array
        The distance from the edge of the rectangle.
    """
    dist_from_vert = np.abs(coords[0]) - width / 2
    dist_from_horz = np.abs(coords[1]) - height / 2
    return np.maximum(dist_from_vert, dist_from_horz)


def spider_distance(coords: Array, width: float, angle: float) -> Array:
    """
    Calculates the distance from the edge of a spider.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the spider.
    width : float
        The width of the spider.
    angle : float
        The angle of the spider.

    Returns
    -------
    distance : Array
        The distance from the edge of the spider.
    """
    coords = dlu.rotate_coords(coords, dlu.deg2rad(angle))
    dist_from_vert = np.abs(coords[0]) - width / 2
    dist_from_horz = coords[1]
    return np.maximum(dist_from_vert, dist_from_horz)


def line_distance(coords: Array, m: float, xy: Array) -> Array:
    """
    Calculates the distance from the edge of a line. This should work for the irregular
    polygon case too, when it is implemented.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the line.
    m : float
        The gradient of the line.
    xy : Array
        The point on the line.

    Returns
    -------
    distance : Array
        The distance from the edge of the line.
    """
    # Extract x and y coordinates from the input array
    x, y = coords

    # First determine the case: horizontal, vertical or regular
    # NOTE: We may be able to improve here my using np.where
    case = (m == 0) * 0 + (m == np.inf) * 1 + (m != 0) * (m != np.inf) * 2

    # Now we determine the sign on the y-intersection so we know when to flip
    # the sign of the distance
    horiz_fn = lambda: np.sign(xy[1])
    vert_fn = lambda: np.sign(xy[0])
    base_fn = lambda: np.sign(xy[1] - m * xy[0])
    sgn = lax.switch(case, [horiz_fn, vert_fn, base_fn])

    # Now we calculate the distance from the point to the line
    horiz_fn = lambda: xy[1] - y
    vert_fn = lambda: xy[0] - x
    base_fn = lambda: (m * x - y + xy[1] - m * xy[0]) / np.sqrt(m**2 + 1)
    dist = lax.switch(case, [horiz_fn, vert_fn, base_fn])

    # Finally, return the distances with the correct sign
    return -sgn * dist


def reg_polygon_edges(n: int, radius: float) -> Array:
    """
    Calculates the gradients and points on the edges of a regular polygon. Can probably
    be extended to irregular polygons too by taking in x, y.

    Parameters
    ----------
    n : int
        The number of sides of the polygon.
    radius : float
        The radius of the polygon.

    Returns
    -------
    m : Array
        The gradients of the edges of the polygon.
    """
    # Calculate the x and y coordinates of the vertices of the polygon
    x = radius * np.cos(np.linspace(-np.pi, np.pi, n, endpoint=False))
    y = radius * np.sin(np.linspace(-np.pi, np.pi, n, endpoint=False))

    # Generate the indexes of the vertices and the next vertex
    ivals = np.arange(n)
    jvals = np.roll(ivals, -1)

    # Calculate the gradient and point on each edge of the polygon
    calc_m = vmap(lambda i, j: dlu.nandiv(y[j] - y[i], x[j] - x[i]), (0, 0))
    calc_xy = vmap(lambda i, j: np.array([x[i], y[i]]), (0, 0))
    return calc_m(ivals, jvals), calc_xy(ivals, jvals)


def reg_polygon_distance(coords: Array, nsides: int, radius: float) -> Array:
    """
    Calculates the distance from the edge of a regular polygon.

    Parameters
    ----------
    coords : Array
        The coordinates to calculate the distance from the edge of the polygon.
    nsides : int
        The number of sides of the polygon.
    radius : float
        The radius of the polygon.

    Returns
    -------
    distance : Array
        The distance from the edge of the polygon.
    """
    m, xy = reg_polygon_edges(nsides, radius)
    return vmap(line_distance, (None, 0, 0))(coords, m, xy).max(0)



from __future__ import annotations
import jax.numpy as np
from jax import Array


from .optical_layers import OpticalLayer
from ..wavefronts import Wavefront


__all__ = ["MFT", "FFT", "ShiftedMFT", "FarFieldFresnel"]


class Propagator(OpticalLayer):
    """
    Base propagator class, instantiating the focal_length attribute.

    Attributes
    ----------
    focal_length : float
        The effective focal length of the focusing optic this class represents. Note if
        `focal_length` is None, pixel_scales are assumed to be in radians/pixel, else
        meters/pixel.
    """

    focal_length: float

    def __init__(self: Propagator, focal_length: float = None):
        """
        Parameters
        ----------
        focal_length : float = None
            The effective focal length of the focusing optic this class represents. Note
            if `focal_length` is None, pixel_scales are assumed to be in radians/pixel,
            else meters/pixel.
        """
        super().__init__()

        if focal_length is not None:
            focal_length = float(focal_length)

        self.focal_length = focal_length


class FFT(Propagator):
    """
    Propagates a `Wavefront` using the FFT algorithm.

    ??? abstract "UML"
        ![UML](../../assets/uml/FFT.png)

    # TODO: Update padding to take in units of npixels, rather than factor.

    Attributes
    ----------
    focal_length : float
        The focal_length of the lens/mirror this propagator represents. If None, the
        output pixel_scale has units radians/pixel, else meters/pixels.
    pad : int
        The zero-padding factor to apply to the `Wavefront` before propagation. In
        general, this should be greater than 2 to avoid aliasing.
    """

    pad: int

    def __init__(self: Propagator, focal_length: float = None, pad: int = 2):
        """
        Parameters
        ----------
        focal_length : float = None
            The focal_length of the lens/mirror this propagator represents. If None, the
            output pixel_scale has units radians/pixel, else meters/pixels.
        pad : int = 2
            The zero-padding factor to apply to the `Wavefront` before propagation. In
            general, this should be greater than 2 to avoid aliasing.
        """
        super().__init__(focal_length=focal_length)
        self.pad = int(pad)

    def apply(self: Propagator, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.propagate_FFT(self.focal_length, self.pad)


class MFT(Propagator):
    """
    Propagates a `Wavefront` using the MFT algorithm, allowing for the pixel_scale and
    number of pixels to be specified in the output plane.

    ??? abstract "UML"
        ![UML](../../assets/uml/MFT.png)

    Attributes
    ----------
    npixels : int
        The number of pixels in the output plane.
    pixel_scale : float, meters/pixel or radians/pixel
        The pixel scale in the output plane. Has units of radians/pixel if focal_length
        is None, else meters/pixel.
    focal_length : float, meters
        The focal_length of the lens/mirror this propagator represents. If None, the
        output pixel_scale has units radians/pixel, else meters/pixels.
    """

    npixels: int
    pixel_scale: float

    def __init__(
        self: Propagator,
        npixels: int,
        pixel_scale: float,
        focal_length: float = None,
    ):
        """
        Parameters
        ----------
        npixels : int
            The number of pixels in the output plane.
        pixel_scale : float, meters/pixel or radians/pixel
            The pixel scale in the output plane. Has units of radians/pixel if
            focal_length is None, else meters/pixel.
        focal_length : float = None, meters
            The focal_length of the lens/mirror this propagator represents. If None,
            the output pixel_scale has units radians/pixel, else meters/pixels.
        """
        super().__init__(focal_length=focal_length)

        self.pixel_scale = float(pixel_scale)
        self.npixels = int(npixels)

    def apply(self: Propagator, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.propagate(
            self.npixels,
            self.pixel_scale,
            self.focal_length,
        )


class ShiftedMFT(MFT):
    """
    Propagates a `Wavefront` using the MFT algorithm, allowing for the pixel_scale and
    number of pixels to be specified in the output plane. Also optionally allows for a
    shift to be applied to the wavefront in the output plane.

    ??? abstract "UML"
        ![UML](../../assets/uml/ShiftedMFT.png)

    Attributes
    ----------
    npixels : int
        The number of pixels in the output plane.
    pixel_scale : float, meters/pixel or radians/pixel
        The pixel scale in the output plane. Has units of radians/pixel if focal_length
        is None, else meters/pixel.
    focal_length : float, meters
        The focal_length of the lens/mirror this propagator represents. If None, the
        output pixel_scale has units radians/pixel, else meters/pixels.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel : bool
        If True the shift value is assumed to be in units of pixels, else the physical
        units of the output plane (ie radians if focal_length is None, else meters).
    """

    shift: Array
    pixel: bool

    def __init__(
        self: Propagator,
        npixels: int,
        pixel_scale: float,
        shift: Array,
        focal_length: float = None,
        pixel: bool = False,
    ):
        """
        Parameters
        ----------
        npixels : int
            The number of pixels in the output plane.
        pixel_scale : float, meters/pixel or radians/pixel
            The pixel scale in the output plane. Has units of radians/pixel if
            focal_length is None, else meters/pixel.
        shift : Array
            The (x, y) shift to apply to the wavefront in the output plane.
        focal_length : float = None, meters
            The focal_length of the lens/mirror this propagator represents. If None,
            the output pixel_scale has units radians/pixel, else meters/pixels.
        pixel : bool = False
            If True the shift value is assumed to be in units of pixels, else the
            physical units of the output plane (ie radians if focal_length is None,
            else meters).
        """
        super().__init__(
            pixel_scale=pixel_scale,
            npixels=npixels,
            focal_length=focal_length,
        )

        self.shift = np.asarray(shift, float)
        self.pixel = bool(pixel)

        if self.shift.shape != (2,):
            raise ValueError(
                f"Shift must be a 2D array, got {self.shift.shape}."
            )

    def apply(self: Propagator, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.propagate(
            self.npixels,
            self.pixel_scale,
            self.focal_length,
            self.shift,
            self.pixel,
        )


class FarFieldFresnel(ShiftedMFT):
    """
    Propagates a `Wavefront` using the MFT algorithm, modified to allows for more
    accurate representations of wavefront behaviour in the far-field regime, a few
    wavelengths away from focus in either direction. Allows for the pixel_scale and
    number of pixels to be specified in the output plane, and optionally allows for a
    shift to be applied to the wavefront in the output plane.

    ??? abstract "UML"
        ![UML](../../assets/uml/FarFieldFresnel.png)

    Attributes
    ----------
    npixels : int
        The number of pixels in the output plane.
    pixel_scale : float, meters/
        The pixel scale in the output plane.
    focal_length : float, meters
        The focal_length of the lens/mirror this propagator represents.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel : bool
        If True the shift value is assumed to be in units of pixels, else the physical
        units of the output plane.
    focal_shift : float, meters
        The shift in the propagation distance of the wavefront from focus.
    """

    focal_shift: float

    def __init__(
        self: Propagator,
        npixels: int,
        pixel_scale: float,
        focal_length: float,
        focal_shift: float,
        shift: Array = np.zeros(2),
        pixel: bool = False,
    ):
        """
        Parameters
        ----------
        npixels : int
            The number of pixels in the output plane.
        pixel_scale : float, meters/
            The pixel scale in the output plane.
        focal_length : float, meters
            The focal_length of the lens/mirror this propagator represents.
        focal_shift : float, meters
            The shift in the propagation distance of the wavefront from focus.
        shift : Array = np.zeros(2)
            The (x, y) shift to apply to the wavefront in the output plane.
        pixel : bool = False
            If True the shift value is assumed to be in units of pixels, else the
            physical units of the output plane.
        """
        self.focal_shift = float(focal_shift)
        super().__init__(
            shift=shift,
            pixel=pixel,
            focal_length=focal_length,
            pixel_scale=pixel_scale,
            npixels=npixels,
        )

    def apply(self: Propagator, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        return wavefront.propagate_fresnel(
            self.npixels,
            self.pixel_scale,
            self.focal_length,
            self.focal_shift,
            self.shift,
            self.pixel,
        )



import sys
import os

platform: str = sys.platform
not_lin_err: str = """
    The installation test is designed to run on a linux machine. In particular,
    this test is designed to be run by github actions.
"""

if platform not in ["linux", "linux2"]:
    print(f"Error: {not_lin_err}")


def test_install_dLux():
    install_ok = os.system("pip install --quiet .")

    assert install_ok == 0

    os.system("rm -r dLux.egg-info")
    os.system("rm -r build")


def test_import_dLux():
    import dLux

    dLux.optical_systems
    assert True



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import (
    Scene,
    PointSource,
    PointSources,
    BinarySource,
    ResolvedSource,
    PointResolvedSource,
    Spectrum,
    LayeredOpticalSystem,
    Wavefront,
    Optic,
    PSF,
)


def _test_model(source):
    optics = LayeredOpticalSystem(16, 1.0, [Optic()])
    assert isinstance(source.model(optics), np.ndarray)
    assert isinstance(source.model(optics, return_wf=True), Wavefront)
    assert isinstance(source.model(optics, return_psf=True), PSF)
    with pytest.raises(ValueError):
        source.model(optics, return_wf=True, return_psf=True)


@pytest.fixture
def wavelengths():
    return np.ones(2)


@pytest.fixture
def spectrum():
    return Spectrum([1e-6])


def test_point_source(wavelengths, spectrum):
    _test_model(PointSource(wavelengths))
    _test_model(PointSource(wavelengths, spectrum=spectrum))

    with pytest.raises(ValueError):
        PointSource(wavelengths, position=np.ones(1))
    with pytest.raises(TypeError):
        PointSource(wavelengths, spectrum=1)


def test_point_sources(wavelengths, spectrum):
    _test_model(PointSources(wavelengths))
    _test_model(PointSources(wavelengths, spectrum=spectrum))
    _test_model(PointSources(wavelengths, flux=np.ones(1)))

    with pytest.raises(ValueError):
        PointSources(wavelengths, position=np.ones(1))
    with pytest.raises(ValueError):
        PointSources(wavelengths, flux=np.ones(3))
    with pytest.raises(ValueError):
        PointSources(wavelengths, flux=np.ones((1, 3)))


def test_resolved_source(wavelengths, spectrum):
    _test_model(ResolvedSource(wavelengths))

    with pytest.raises(ValueError):
        ResolvedSource(wavelengths, distribution=np.ones(1))


def test_binary_source(wavelengths, spectrum):
    _test_model(BinarySource(wavelengths, separation=1.0))

    with pytest.raises(ValueError):
        BinarySource(wavelengths, separation=1.0, position=np.ones(1))


def test_point_resolved_source(wavelengths, spectrum):
    _test_model(PointResolvedSource(wavelengths))


def test_scene():
    scene = Scene(("source", PointSource([1e-6])))
    scene = Scene([PointSource([1e-6]), PointSource([1e-6])])
    optics = LayeredOpticalSystem(16, 1.0, [Optic()])

    # In this case we have an output dictionary of each source
    assert isinstance(scene.model(optics), np.ndarray)
    output = scene.model(optics, return_wf=True).values()
    for o in output:
        assert isinstance(o, Wavefront)

    # In this case we have a single PSF object
    assert isinstance(scene.model(optics, return_psf=True), PSF)

    # Test getattr
    scene.PointSource_0
    with pytest.raises(AttributeError):
        scene.not_an_attr



import importlib.metadata

__version__ = importlib.metadata.version("dLux")

from . import (
    layers,
    optical_systems,
    detectors,
    instruments,
    sources,
    spectra,
    wavefronts,
    psfs,
    transformations,
)

# Add to __all__
modules = [
    optical_systems,
    detectors,
    sources,
    spectra,
    instruments,
    layers,
    wavefronts,
    psfs,
    transformations,
]

__all__ = [module.__all__ for module in modules]


from .detectors import (
    BaseDetector as BaseDetector,
    LayeredDetector as LayeredDetector,
)
from .spectra import (
    BaseSpectrum as BaseSpectrum,
    Spectrum as Spectrum,
    PolySpectrum as PolySpectrum,
)
from .wavefronts import Wavefront as Wavefront
from .psfs import PSF as PSF
from .transformations import CoordTransform as CoordTransform

from .instruments import (
    Instrument as Instrument,
    Telescope as Telescope,
    Dither as Dither,
)

from .optical_systems import (
    BaseOpticalSystem as BaseOpticalSystem,
    AngularOpticalSystem as AngularOpticalSystem,
    CartesianOpticalSystem as CartesianOpticalSystem,
    LayeredOpticalSystem as LayeredOpticalSystem,
)

from .sources import (
    BaseSource as BaseSource,
    Scene as Scene,
    PointSource as PointSource,
    PointSources as PointSources,
    BinarySource as BinarySource,
    ResolvedSource as ResolvedSource,
    PointResolvedSource as PointResolvedSource,
)

from .layers import (
    optical_layers as optical_layers,
    detector_layers as detector_layers,
    unified_layers as unified_layers,
    propagators as propagators,
    apertures as apertures,
    aberrations as aberrations,
)

# Import core functions from modules
from .layers.optical_layers import (
    BaseLayer as BaseLayer,
    TransmissiveLayer as TransmissiveLayer,
    AberratedLayer as AberratedLayer,
    BasisLayer as BasisLayer,
    Tilt as Tilt,
    Normalise as Normalise,
)
from .layers.optics import (
    Optic as Optic,
    BasisOptic as BasisOptic,
)
from .layers.detector_layers import (
    ApplyPixelResponse as ApplyPixelResponse,
    ApplyJitter as ApplyJitter,
    ApplySaturation as ApplySaturation,
    AddConstant as AddConstant,
    Downsample as Downsample,
)
from .layers.unified_layers import (
    Rotate as Rotate,
    Flip as Flip,
    Resize as Resize,
)
from .layers.propagators import (
    MFT as MFT,
    FFT as FFT,
    ShiftedMFT as ShiftedMFT,
    FarFieldFresnel as FarFieldFresnel,
)
from .layers.apertures import (
    CircularAperture as CircularAperture,
    SquareAperture as SquareAperture,
    RectangularAperture as RectangularAperture,
    RegPolyAperture as RegPolyAperture,
    # IrregPolyAperture as IrregPolyAperture,
    Spider as Spider,
    AberratedAperture as AberratedAperture,
    CompoundAperture as CompoundAperture,
    MultiAperture as MultiAperture,
    # ApertureFactory as ApertureFactory,
)
from .layers.aberrations import (
    Zernike as Zernike,
    ZernikeBasis as ZernikeBasis,
)



import jax.numpy as np
from jax import Array

__all__ = [
    "fluxes_from_contrast",
    "positions_from_sep",
]


def fluxes_from_contrast(mean_flux: float, contrast: float) -> Array:
    """
    Computes the fluxes of a binary object given the mean flux and contrast.

    Parameters
    ----------
    mean_flux : float
        The mean flux of the binary object.
    contrast : float
        The contrast of the binary object.

    Returns
    -------
    fluxes : Array
        The flux (flux1, flux2) of the binary object.
    """
    return 2 * np.array([contrast * mean_flux, mean_flux]) / (1 + contrast)


def positions_from_sep(
    position: Array, separation: float, position_angle: float
) -> Array:
    """
    Computes the on-sky positions of a binary object given the separation and
    position angle.

    Parameters
    ----------
    position : Array, radians
        The on-sky position of the primary object.
    separation : float, radians
        The separation of the binary object.
    position_angle : float, radians
        The position angle of the binary object.

    Returns
    -------
    position : Array, radians
        The ((x, y), (x, y)) on-sky position of this object.
    """
    r, phi = separation / 2, position_angle
    sep_vec = np.array([r * np.sin(phi), r * np.cos(phi)])
    return np.array([position + sep_vec, position - sep_vec])



from __future__ import annotations
import jax.numpy as np
from jax import lax, Array
import dLux.utils as dlu

__all__ = [
    "zernike_name",
    "noll_indices",
    "zernike_factors",
    "eval_radial",
    "eval_azimuthal",
    "zernike",
    "zernike_fast",
    "zernike_basis",
    "polike",
    "polike_fast",
    "polike_basis",
]

zernike_names = {
    # 0th Radial
    1: "Piston",
    # 1st Radial
    2: "Tilt X",
    3: "Tilt Y",
    # Second Radial
    4: "Defocus",
    5: "Astig X",
    6: "Astig Y",
    # Third Radial
    7: "Coma X",
    8: "Coma Y",
    9: "Trefoil X",
    10: "Trefoil Y",
    # Fourth Radial
    11: "Spherical",
    12: "2nd Astig X",
    13: "2nd Astig Y",
    14: "Quadrafoil X",
    15: "Quadrafoil Y",
    # Fifth Radial
    16: "2nd Coma X",
    17: "2nd Coma Y",
    18: "2nd Trefoil X",
    19: "2nd Trefoil Y",
    20: "Pentafoil X",
    21: "Pentafoil Y",
    # Sixth Radial
    22: "2nd Spherical",
    23: "3rd Coma X",
    24: "3rd Coma Y",
    25: "3rd Astig X",
    26: "3rd Astig Y",
    27: "Hexafoil X",
    28: "Hexafoil Y",
    # Seventh Radial
    29: "4th Coma X",
    30: "4th Coma Y",
    31: "4th Astig X",
    32: "4th Astig Y",
    33: "3rd Trefoil X",
    34: "3rd Trefoil Y",
    35: "Heptafoil X",
    36: "Heptafoil Y",
}


def zernike_name(j: int) -> str:
    """
    Gets the name of the jth Zernike polynomial.

    Parameters
    ----------
    j : int
        The Zernike (noll) index.

    Returns
    -------
    name : str
        The name of the Zernike polynomial.
    """
    return zernike_names[int(j)] if j >= 1 and j <= 36 else f"Zernike {int(j)}"


def noll_indices(j: int) -> tuple[int]:
    """
    Calculate the radial and azimuthal orders of the Zernike polynomial.

    Parameters
    ----------
    j : int
        The Zernike (noll) index.

    Returns
    -------
    n, m : tuple[int]
        The radial and azimuthal orders of the Zernike polynomial.
    """
    n = (np.ceil(-1 / 2 + np.sqrt(1 + 8 * j) / 2) - 1).astype(int)
    smallest_j_in_row = n * (n + 1) / 2 + 1
    number_of_shifts = (j - smallest_j_in_row + ~(n & 1) + 2) // 2
    sign_of_shift = -(j & 1) + ~(j & 1) + 2
    base_case = n & 1
    m = (sign_of_shift * (base_case + number_of_shifts * 2)).astype(int)
    return int(n), int(m)


def zernike_factors(j: int) -> tuple[Array]:
    """
    Calculates the normalisation coefficients and powers of the Zernike polynomial.

    Parameters
    ----------
    j : int
        The Zernike (noll) index.

    Returns
    -------
    c, k : tuple[Array]
        The normalisation coefficients and powers of the Zernike polynomial.
    """
    if j < 1:
        raise ValueError("The Zernike index must be greater than 0.")
    n, m = noll_indices(j)

    # Calculate k
    k = np.arange(((n - m) // 2) + 1, dtype=float)

    # Calculate c
    sign = lax.pow(-1.0, k)
    _fact_1 = dlu.factorial(np.abs(n - k))
    _fact_2 = dlu.factorial(k)
    _fact_3 = dlu.factorial(((n + m) // 2) - k)
    _fact_4 = dlu.factorial(((n - m) // 2) - k)
    c = sign * _fact_1 / _fact_2 / _fact_3 / _fact_4
    return c, k


def eval_radial(rho: Array, n: int, c: Array, k: Array) -> Array:
    """
    Calculates the radial component of the Zernike polynomial.

    Parameters
    ----------
    rho : Array
        The radial coordinate of the Zernike polynomial.
    n : int
        The radial order of the Zernike polynomial.
    c : Array
        The normalisation coefficients of the Zernike polynomial.
    k : Array
        The powers of the Zernike polynomial.

    Returns
    -------
    radial : Array
        The radial component of the Zernike polynomial.
    """
    rads = lax.pow(rho[:, :, None], (np.abs(n) - 2 * k)[None, None, :])
    return (c * rads).sum(axis=2)


def eval_azimuthal(theta: Array, n: int, m: int) -> Array:
    """
    Calculates the azimuthal component of the Zernike polynomial.

    Parameters
    ----------
    theta : Array
        The azimuthal coordinate of the Zernike polynomial.
    n : int
        The radial order of the Zernike polynomial.
    m : int
        The azimuthal order of the Zernike polynomial.

    Returns
    -------
    azimuthal : Array
        The azimuthal component of the Zernike polynomial.
    """
    # Get normalisation coefficient
    norm_coeff = np.sqrt(n + 1)
    if m != 0:
        norm_coeff *= 1 + (np.sqrt(2) - 1)

    # Get the right trig function and eval
    trig_fn = np.cos if m >= 0 else np.sin
    return norm_coeff * trig_fn(np.abs(m) * theta)


def scale_coords(coords: Array, rmax: float) -> Array:
    """
    Scales coordinates to the unit circle, to calculate polynomials on a subset of the
    coordinates.

    Parameters
    ----------
    coords : Array
        The coordinates to scale.
    rmax : float
        The radius to scale to.

    Returns
    -------
    coords : Array
        The scaled coordinates.
    """
    return coords / rmax


def zernike(j: int, coordinates: Array, diameter: float = 2) -> Array:
    """
    Calculates the Zernike polynomial. Note that this function is not-jittable as is
    has dynamic array shapes. To use this function in a jittable way, use the
    zernike_fast function, with the pre-calculated c and k parameters.

    Parameters
    ----------
    j : int
        The Zernike (noll) index.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.
    diameter : float = 2
        The diameter of the aperture to calculate the Zernike polynomial upon.

    Returns
    -------
    zernike : Array
        The Zernike polynomial.
    """
    coordinates = scale_coords(coordinates, diameter / 2)
    polar_coordinates = dlu.cart2polar(coordinates)
    rho = polar_coordinates[0]
    theta = polar_coordinates[1]
    aperture = rho <= 1.0
    n, m = noll_indices(j)
    c, k = zernike_factors(j)
    return aperture * eval_radial(rho, n, c, k) * eval_azimuthal(theta, n, m)


def zernike_fast(
    n: int, m: int, c: Array, k: Array, coordinates: Array
) -> Array:
    """
    Calculates the Zernike polynomial using the pre-calculated c and k parameters, such
    that this function is jittable.

    Parameters
    ----------
    n : int
        The radial order of the Zernike polynomial.
    m : int
        The azimuthal order of the Zernike polynomial.
    c : Array
        The normalisation coefficients of the Zernike polynomial.
    k : Array
        The powers of the Zernike polynomial.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.

    Returns
    -------
    zernike : Array
        The Zernike polynomial.
    """
    polar_coordinates = dlu.cart2polar(coordinates)
    rho = polar_coordinates[0]
    theta = polar_coordinates[1]
    aperture = rho <= 1.0
    return aperture * eval_radial(rho, n, c, k) * eval_azimuthal(theta, n, m)


def zernike_basis(
    js: list[int], coordinates: Array, diameter: float = 2
) -> Array:
    """
    Calculates the Zernike polynomial basis. Note that this function is not-jittable.

    Parameters
    ----------
    js : list[int]
        The Zernike (noll) indices.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.
    diameter : float = 2
        The diameter of the aperture to calculate the Zernike polynomial upon.

    Returns
    -------
    zernike_basis : Array
        The Zernike polynomial basis.
    """
    return np.array([zernike(j, coordinates, diameter) for j in js])


def polike(
    nsides: int, j: int, coordinates: Array, diameter: float = 2
) -> Array:
    """
    Calculates the Zernike polynomial on an n-sided aperture. Note that this function
    is not-jittable as is has dynamic array shapes. To use this function in a jittable
    way, use the polike_fast function, with the pre-calculated c and k parameters.

    Parameters
    ----------
    nsides : int
        The number of sides of the aperture.
    j : int
        The Zernike (noll) index.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.
    diameter : float = 2
        The diameter of the aperture to calculate the Zernike polynomial upon.

    Returns
    -------
    polike : Array
        The Zernike polynomial on an n-sided aperture.
    """
    if nsides < 3:
        raise ValueError(f"nsides must be >= 3, not {nsides}.")
    coordinates = scale_coords(coordinates, diameter / 2)
    theta = dlu.cart2polar(coordinates)[1]
    alpha = np.pi / nsides
    phi = theta + alpha
    wedge = np.floor((phi + alpha) / (2.0 * alpha))
    u_alpha = phi - wedge * (2 * alpha)
    r_alpha = np.cos(alpha) / np.cos(u_alpha)
    return 1 / r_alpha * zernike(j, coordinates / r_alpha)


def polike_fast(
    nsides: int, n: int, m: int, c: Array, k: Array, coordinates: Array
) -> Array:
    """
    Calculates the Zernike polynomial on an n-sided aperture using the pre-calculated
    c and k parameters, such that this function is jittable.

    Parameters
    ----------
    nsides : int
        The number of sides of the aperture.
    n : int
        The radial order of the Zernike polynomial.
    m : int
        The azimuthal order of the Zernike polynomial.
    c : Array
        The normalisation coefficients of the Zernike polynomial.
    k : Array
        The powers of the Zernike polynomial.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.

    Returns
    -------
    polike : Array
        The Zernike polynomial on an n-sided aperture.
    """
    if nsides < 3:
        raise ValueError(f"nsides must be >= 3, not {nsides}.")
    alpha = np.pi / nsides
    phi = dlu.cart2polar(coordinates)[1] + alpha
    wedge = np.floor((phi + alpha) / (2.0 * alpha))
    u_alpha = phi - wedge * (2 * alpha)
    r_alpha = np.cos(alpha) / np.cos(u_alpha)
    return 1 / r_alpha * zernike_fast(n, m, c, k, coordinates / r_alpha)


def polike_basis(
    nsides: int, js: list[int], coordinates: Array, diameter: float = 2
):
    """
    Calculates the Zernike polynomial basis on an n-sided aperture. Note that this
    function is not-jittable.

    Parameters
    ----------
    nsides : int
        The number of sides of the aperture.
    js : list[int]
        The Zernike (noll) indices.
    coordinates : Array
        The Cartesian coordinates to calculate the Zernike polynomial upon.
    diameter : float = 2
        The diameter of the aperture to calculate the Zernike polynomial upon.

    Returns
    -------
    polike_basis : Array
        The Zernike polynomial basis on an n-sided aperture.
    """
    return np.array([polike(nsides, j, coordinates, diameter) for j in js])
